# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

# VV: @actuator_version Update this version according to semver.
# Consider "refactoring" the code as a "bug patch".
# Use `git-bisect` when looking for the implementation of a specific actuator version.
# Also remember to update the experiment descriptions in the `README.md` as bumping the version of the actuator
# also bumps the version of *all* currently supported experiments.
import copy
import enum
import functools
import logging
import os
import typing

import ado_actuators.sfttrainer.wrapper_fms_hf_tuning.constants as constants
import pydantic
import pydantic.fields
import pydantic_core
import yaml

import orchestrator.schema.domain
import orchestrator.schema.property_value
from orchestrator.schema.domain import PropertyDomain, VariableTypeEnum
from orchestrator.schema.experiment import Experiment
from orchestrator.schema.property import (
    ConstitutiveProperty,
    ConstitutivePropertyDescriptor,
)
from orchestrator.schema.property_value import ConstitutivePropertyValue


class InternalInconsistencyError(ValueError):
    pass


class InvalidEntityError(ValueError):
    pass


ACTUATOR_IDENTIFIER = "SFTTrainer"
ACTUATOR_VERSION = "2.2.0"

FMS_HF_TUNING_REPOSITORY = "https://github.com/foundation-model-stack/fms-hf-tuning"
PACKAGES_DIR = f"{os.path.dirname(__file__)}/../packages"
CONFIG_DIR = f"{os.path.dirname(__file__)}/../config"

with open(os.path.join(CONFIG_DIR, "map_version_to_commit.yaml")) as f:
    FMS_HF_TUNING_COMMIT = yaml.safe_load(f)

FMS_HF_TUNING_VERSION = {
    version: f"{FMS_HF_TUNING_REPOSITORY}/tree/{commit}"
    for version, commit in FMS_HF_TUNING_COMMIT.items()
}


PATH_PINNED_PACKAGES = {
    version: f"{PACKAGES_DIR}/fms-hf-tuning_v{version}_{commit}.txt"
    for version, commit in FMS_HF_TUNING_COMMIT.items()
}


BACKEND_NAME_MAP = {
    "FSDP": "fsdp",
    "DDP": "ddp",
    None: "dp",
}


VALUES_TRANSFORMERS_ARGUMENT_OPTIM = [
    "adamw_hf",
    "adamw_torch",  # VV: transformers uses this as the default value of the --optim arg
    "adamw_torch_fused",
    "adamw_torch_xla",
    "adamw_torch_npu_fused",
    "adamw_apex_fused",
    "adafactor",
    "adamw_anyprecision",
    "adamw_torch_4bit",
    "ademamix",
    "sgd",
    "adagrad",
    "adamw_bnb_8bit",
    "adamw_8bit",
    "ademamix_8bit",
    "lion_8bit",
    "lion_32bit",
    "paged_adamw_32bit",  # VV: The blog-post references this one
    "paged_adamw_8bit",
    "paged_ademamix_32bit",
    "paged_ademamix_8bit",
    "paged_lion_32bit",
    "paged_lion_8bit",
    "rmsprop",
    "rmsprop_bnb",
    "rmsprop_bnb_8bit",
    "rmsprop_bnb_32bit",
    "galore_adamw",
    "galore_adamw_8bit",
    "galore_adafactor",
    "galore_adamw_layerwise",
    "galore_adamw_8bit_layerwise",
    "galore_adafactor_layerwise",
    "lomo",
    "adalomo",
    "grokadamw",
    "schedule_free_adamw",
    "schedule_free_sgd",
]


def packages_requiring_nvidia_development_binaries():
    return [
        "fms-acceleration-foak",
        "fms-acceleration-moe",
        "triton",
        "flash_attn",
        "mamba-ssm",
        "causal-conv1d",
        "nvidia-cublas-cu12",
        "nvidia-cuda-cupti-cu12",
        "nvidia-cuda-nvrtc-cu12",
        "nvidia-cuda-runtime-cu12",
        "nvidia-cudnn-cu12",
        "nvidia-cufft-cu12",
        "nvidia-curand-cu12",
        "nvidia-cusolver-cu12",
        "nvidia-cusparse-cu12",
        "nvidia-nccl-cu12",
        "nvidia-nvjitlink-cu12",
        "nvidia-nvtx-cu12",
    ]


def parse_semver(what: str) -> list | None:
    import re

    p = re.compile(r"(\d+)\.(\d+)\.(\d+)")
    m = p.fullmatch(what)

    if m is None:
        return None

    return [int(x) for x in [m.group(1), m.group(2), m.group(3)]]


semvers = [x for x in map(parse_semver, FMS_HF_TUNING_VERSION) if x is not None]

# VV: We use these to instantiate Optional Parameters for all experiments (values are the defaults of the params)


class WeightsFormat(str, enum.Enum):
    Vanilla = "Vanilla"
    GPTQQuantized = "GPTQ-Quantized"

    def __str__(self):
        return self.value


class ExperimentPurpose(str, enum.Enum):
    Performance = "Performance"
    Stability = "Stability"

    def __str__(self):
        return self.value


@functools.cache
def load_model_map() -> dict[str, dict[WeightsFormat, str]]:
    from importlib.resources import read_text

    return yaml.safe_load(read_text("ado_actuators.sfttrainer.config", "models.yaml"))


def get_default_measured_properties() -> list[str]:
    return [
        "is_valid",
        "dataset_tokens_per_second_per_gpu",
        "train_runtime",
        "dataset_tokens_per_second",
        # VV: the next 4 are inaccurate when terminating the job early
        "train_samples_per_second",
        "train_steps_per_second",
        "train_tokens_per_second",
        "train_tokens_per_gpu_per_second",
        # VV: We no longer record the model_load_time
        # "model_load_time",
        # VV: CPU measurements
        "cpu_compute_utilization",
        "cpu_memory_utilization",
        # VV: GPU measurements - these will be 0 when running on a machine without GPUs
        "gpu_compute_utilization_min",
        "gpu_compute_utilization_avg",
        "gpu_compute_utilization_max",
        "gpu_memory_utilization_min",
        "gpu_memory_utilization_avg",
        "gpu_memory_utilization_max",
        "gpu_memory_utilization_peak",
        "gpu_power_watts_min",
        "gpu_power_watts_avg",
        "gpu_power_watts_max",
        "gpu_power_percent_min",
        "gpu_power_percent_avg",
        "gpu_power_percent_max",
    ]


ModelMap = load_model_map()

# VV: by convention, we use .jsonl for text datasets
DatasetMap = {
    x: f"{x}.jsonl"
    for x in (
        # VV: entries = max_batch_size * gradient_accumulation = 8 * 128 * 1
        "news-chars-512-entries-1024",
        "news-chars-1024-entries-1024",
        "news-chars-2048-entries-1024",
        # VV: entries = max_batch_size * gradient_accumulation = 8 * 128 * 4
        "news-chars-512-entries-4096",
        "news-chars-1024-entries-4096",
        "news-chars-2048-entries-4096",
        "news-tokens-16384plus-entries-4096",
        "news-tokens-128kplus-entries-4096",
        # VV: entries = max_batch_size * gradient_accumulation = 2 * 32 * 4
        "news-chars-512-entries-256",
        "news-chars-1024-entries-256",
        "news-chars-2048-entries-256",
        # SV: entries = max_batch_size * gradient_accumulation = 8 * 10 * 4
        "news-tokens-128kplus-entries-320",
    )
}


# VV: by convention, we use .parquet for vision datasets
DatasetMap.update(
    {
        x: f"{x}.parquet"
        for x in (
            "vision-384x384-16384plus-entries-4096",
            "vision-384x768-16384plus-entries-4096",
        )
    }
)


def get_fms_hf_tuning_package(commit: str) -> str:
    return f"fms-hf-tuning@git+{FMS_HF_TUNING_REPOSITORY}@{commit}"


def get_fms_hf_tuning_src(fms_hf_tuning_version: str) -> str:
    # VV: TODO: Generate links to pypi if fms_hf_tuning_version is a valid pypi version
    return FMS_HF_TUNING_VERSION[fms_hf_tuning_version]


def get_versioning_metadata(
    fms_hf_tuning_version: str | list[str],
) -> dict[str, str]:
    ret = {
        "actuator": ACTUATOR_VERSION,
    }
    if isinstance(fms_hf_tuning_version, str):
        ret.update(
            {
                "version": fms_hf_tuning_version,
                "src": get_fms_hf_tuning_src(
                    fms_hf_tuning_version=fms_hf_tuning_version
                ),
            }
        )
    else:
        ret.update(
            {
                "supported_fms_hf_tuning_versions": [
                    {
                        "fms-hf-tuning-src": get_fms_hf_tuning_src(
                            fms_hf_tuning_version=v
                        ),
                        "fms-hf-tuning": v,
                    }
                    for v in fms_hf_tuning_version
                ]
            }
        )

    return ret


def generate_parameterisable_finetune_experiment(
    description: str,
    method: str,
    exp_name: str,
    exp_identifier: str,
    version: str,
    actuator_identifier: str,
    override_propertydomains: dict[str, orchestrator.schema.domain],
    required_property_names: list[str],
    default_params: dict[str, str | float | bool | int],
    hardcoded_parameters: dict[str, typing.Any],
    fms_hf_tuning_versions: list[str] | str,
    properties: list[str] = ...,
) -> "Experiment":
    """Generates a finetune experiment

    Args:
        description:
            human readable description
        method:
            the finetune method
        exp_name:
            the name of the experiment (basically the identifier but without the version suffix)
        exp_identifier:
            the unique identifier of this experiment
        version:
            the experiment version (e.g. "1.1.1")
        actuator_identifier:
            the actuator identifier
        override_propertydomains:
            A dictionary whose keys are property names for which to use the value as the propertyDomain instead of
            the propertyDomain returned by EntitySpace.domain_for_constitutive_property().
            Use this for experiments which have their own custom rules for propertyDomain
        required_property_names:
            names of the required properties
        default_params:
            Key value pairs with default values of optional properties for experiment measurements
        fms_hf_tuning_versions:
            A list of supported versions of the fms-hf-tuning framework
        hardcoded_parameters:
            These are hardcoded parameters in the experiment which a user cannot change in any way
        properties:
            the names of the properties that this experiment measures. When not provided they default to
            the output of get_default_measured_properties()

    Returns:
        An experiment
    """
    import pydantic_core

    default_default_params = {
        "max_steps": -1.0,
        "num_train_epochs": 1.0,
        "stop_after_seconds": -1.0,
    }

    default_default_params = {
        k: v for k, v in default_default_params.items() if k not in hardcoded_parameters
    }
    default_default_params.update(default_params)
    default_params = default_default_params

    default_override_property_domains = {
        "max_steps": PropertyDomain(
            domainRange=[-1, 10000 + 1],
            interval=1,
            variableType=VariableTypeEnum.DISCRETE_VARIABLE_TYPE,
        ),
        "num_train_epochs": PropertyDomain(
            domainRange=[1.0, 10000.0 + 1.0],
            interval=1,
            variableType=VariableTypeEnum.DISCRETE_VARIABLE_TYPE,
        ),
        "stop_after_seconds": PropertyDomain(
            domainRange=[-1.0, 100_0000 + 1.0],
            interval=1,
            variableType=VariableTypeEnum.DISCRETE_VARIABLE_TYPE,
        ),
    }

    # VV: fill in common property domains
    default_override_property_domains = {
        k: v
        for k, v in default_override_property_domains.items()
        if k not in hardcoded_parameters
    }

    default_override_property_domains.update(override_propertydomains)
    override_propertydomains = default_override_property_domains

    def property_domain_for_prop(
        identifier: str,
    ) -> PropertyDomain:
        if identifier in override_propertydomains:
            return override_propertydomains[identifier]
        return EntitySpace.domain_for_constitutive_property(
            identifier, EntitySpace.model_fields[identifier]
        )

    optional_properties = [
        ConstitutiveProperty(
            identifier=identifier,
            propertyDomain=property_domain_for_prop(identifier),
            metadata={"description": EntitySpace.model_fields[identifier].description},
        )
        for identifier in set(default_params).union(override_propertydomains)
    ]

    optional_properties = sorted(optional_properties, key=lambda x: x.identifier)

    default_parameterisation = []

    default_parameterisation.extend(
        [
            ConstitutivePropertyValue(
                property=ConstitutivePropertyDescriptor(identifier=identifier),
                value=value,
            )
            for identifier, value in default_params.items()
        ]
    )

    required_properties = [
        p
        for p in EntitySpace.orch_experiment_required_properties()
        if p.identifier in required_property_names
    ]

    req_prop_identifiers = [
        p.identifier
        for p in required_properties
        # VV: All fields of the EntitySpace basemodel represent required properties, except for those which are:
        # 1. hardcoded by this experiment (e.g. a lora tuning method), or
        # 2. parametrisable by this experiment
        if (
            p.identifier not in hardcoded_parameters
            and p.identifier not in default_params
        )
    ]

    required_properties = [
        p for p in required_properties if p.identifier in req_prop_identifiers
    ]

    # VV: Sanity check rules:
    # 1. optional and required properties must **not** include any of the hard-coded params (i.e. base_params)
    # 2. optional properties + required properties + hard coded params = superset all fields of EntitySpace

    property_names = [p.identifier for p in required_properties] + [
        p.identifier for p in optional_properties
    ]
    cannot_be_hardcoded = {
        identifier
        for identifier in hardcoded_parameters
        if identifier in property_names
    }
    if cannot_be_hardcoded:
        raise ValueError(
            "The following hard coded parameters are also properties",
            cannot_be_hardcoded,
        )

    superset = property_names + list(hardcoded_parameters)
    missing = set(EntitySpace.model_fields).difference(superset)

    missing = {
        p
        for p in missing
        if isinstance(
            EntitySpace.model_fields[p].default,
            pydantic_core._pydantic_core.PydanticUndefinedType,
        )
    }

    if missing:
        raise ValueError(
            "The EntitySpace contains the following properties for which there "
            "is neither a property nor a hardcoded parameter",
            missing,
            {EntitySpace.model_fields[p].default for p in missing},
            {
                isinstance(
                    EntitySpace.model_fields[p].default,
                    pydantic_core._pydantic_core.PydanticUndefinedType,
                )
                for p in missing
            },
        )

    if properties is ...:
        properties = get_default_measured_properties()

    versioning = get_versioning_metadata(fms_hf_tuning_version=fms_hf_tuning_versions)
    versioning["experiment"] = version

    is_true_parameterised = False

    for p in optional_properties:
        prop_domain = property_domain_for_prop(p.identifier)

        # VV: An experiment is a true parameterised experimenti if it has at least 1 optional property which
        # can have multiple values
        if (
            prop_domain.domainRange
            and (prop_domain.domainRange[1] - prop_domain.domainRange[0] > 1)
        ) or (prop_domain.values is not None and len(prop_domain.values) > 1):
            is_true_parameterised = True
            break

    # exp_identifier = "-v".join((exp_name, version))
    return Experiment(
        identifier=exp_identifier,
        actuatorIdentifier=actuator_identifier,
        metadata={
            "description": description,
            "method": method,
            "experiment": exp_name,
            "deprecated": False,
            "parameterised": is_true_parameterised,
            "parameters": copy.deepcopy(hardcoded_parameters),
            # VV: See @actuator_version
            "versioning": versioning,
        },
        targetProperties=[
            orchestrator.schema.property.AbstractPropertyDescriptor(identifier=prop)
            for prop in properties
        ],
        optionalProperties=optional_properties,
        requiredProperties=required_properties,
        defaultParameterization=default_parameterisation,
    )


class SFTTrainerCLIArgs(pydantic.BaseModel):
    """These are Entity properties which map to a CLI arg"""

    model_config = pydantic.ConfigDict(
        extra="forbid", protected_namespaces=(), use_enum_values=True
    )

    # VV: If you're updating these, then make sure you also update domain_for_constitutive_property()
    # the code uses `examples` to populate the categorical values of the constitutive property's domain
    model_name: str = pydantic.Field(
        examples=sorted(ModelMap.keys()),
        description="The huggingface name or path to the model",
    )
    model_max_length: int = pydantic.Field(
        examples=[512, 2048, 8192],
        description="The maximum context size. Dataset entries with more tokens they are truncated. Entries with "
        "fewer are padded",
    )

    dataset_id: str = pydantic.Field(
        default="news-tokens-16384plus-entries-4096",
        examples=sorted(DatasetMap.keys()),
        description="The identifier of the dataset to use for training",
    )

    batch_size: int = pydantic.Field(
        examples=[1, 16, 128], description="The total batch size to use"
    )

    # VV: These represent optional properties
    peft_method: typing.Literal["pt", "lora"] | None = pydantic.Field(
        # VV: We auto convert `"full"` into `None`
        examples=["full", "lora", "pt"],
        description='The tuning method. Set to "full" to perform full finetuning',
    )

    gradient_checkpointing: bool = pydantic.Field(
        default=True,
        examples=[True, False],
        description="If True, use gradient checkpointing to save memory (i.e. higher batchsizes) at the expense "
        "of slower backward pass",
    )

    torch_dtype: str = pydantic.Field(
        default="bfloat16",
        examples=["bfloat16", "float16", "float32"],
        description="The torch datatype to use",
    )

    gradient_accumulation_steps: int = pydantic.Field(
        default=4,
        examples=[4],
        description="Number of update steps to accumulate before performing a backward/update pass.",
    )

    # VV: tuning.sft_trainer interprets max_steps and num_train_epochs together
    max_steps: int = pydantic.Field(
        -1,
        examples=[-1, 5],
        description="The number of optimization steps to perform. Set to -1 to respect num_train_epochs instead",
    )
    num_train_epochs: float = pydantic.Field(
        1.0,
        examples=[1.0],
        description="How many epochs to run. Ignored if max_steps is greater than 0",
    )
    stop_after_seconds: float = pydantic.Field(
        default=-1.0,
        examples=[-1, 60 * 10],
        description="If set, the optimizer will be asked to stop after the specified time elapses. "
        "The check is performed after the end of each training step.",
    )

    auto_stop_method: constants.AutoStopMethod | None = pydantic.Field(
        default=None,
        examples=[
            constants.AutoStopMethod.WARMUP_60S_STABLE_120S_OR_10_STEPS.value,
            None,
        ],
        description="The default value is `None`. This parameter defines the method used to automatically "
        "stop the fine-tuning job. Supported values are `WARMUP_60S_STABLE_120S_OR_10_STEPS` and "
        "`None`. If set to `WARMUP_60S_STABLE_120S_OR_10_STEPS`, the job stops after spending at least "
        "60 seconds in the warmup phase plus the longer of 120 seconds or the duration of 10 "
        "optimization steps. This method excludes the first 60 seconds of training when calculating "
        "throughput and system metrics.",
    )

    # VV: lora specific parameters
    r: int = pydantic.Field(4, examples=[4, 8, 16], description="The LORA rank")

    lora_alpha: int = pydantic.Field(
        16, examples=[16], description="LORA Alpha scales the learning weights"
    )

    fast_moe: list[int] | None = pydantic.Field(
        # VV: Here "0" is a stand in for "disabled" -> we translate this into a None via pydantic
        default=0,
        examples=[0, 1, 2, 4, 8],
        description="Configures the amount of expert parallel sharding. number_gpus must be divisible by it",
    )

    fast_kernels: list[str] | None = pydantic.Field(
        default=None,
        description="Switches on fast kernels, the value is a list with strings of boolean values for "
        "[fast_loss, fast_rms_layernorm, fast_rope_embeddings]",
        examples=[None, ["True", "True", "True"]],
    )

    # VV: TAG: @HF_RAM_Efficient_Training
    # VV: These are all parameters that the huggingface blogpost on RAM efficient training uses
    # url: https://huggingface.co/blog/ram-efficient-pytorch-fsdp
    # The default values are all set to those we used for the default finetune v1.0.0 experiment i.e.
    # transformers v4.45.2 as that's the version fms-hf-tuning v2.1.2 uses

    optim: str = pydantic.Field(
        default="adamw_torch",
        description="The optimizer to use.",
        examples=VALUES_TRANSFORMERS_ARGUMENT_OPTIM,
    )

    bf16: bool = pydantic.Field(
        default=False,
        description="Whether to use bf16 (mixed) precision instead of 32-bit. "
        "Requires Ampere or higher NVIDIA add bf16 mixed precision support for NPU "
        "architecture or using CPU (use_cpu) or Ascend NPU. This is an experimental API and it may change.",
        examples=[False, True],
    )

    gradient_checkpointing_use_reentrant: bool = pydantic.Field(
        default=False,
        description="Specify whether to use the activation checkpoint variant that requires reentrant autograd. "
        "This parameter should be passed explicitly. Torch version 2.5 will raise an exception "
        "if use_reentrant is not passed. If use_reentrant=False, checkpoint will use an implementation "
        "that does not require reentrant autograd. This allows checkpoint to support additional functionality, "
        "such as working as expected with torch.autograd.grad and support for keyword arguments input "
        "into the checkpointed function.",
        examples=[False, True],
    )

    # VV: for image-to-text (vision) models
    dataset_text_field: str | None = pydantic.Field(
        default="output",
        examples=["output", "messages"],
        description="Training dataset text field containing single sequence. "
        "Either the dataset_text_field "
        "or data_formatter_template need to be supplied. "
        "For running vision language model tuning pass the column name for text data.",
    )

    dataset_image_field: str | None = pydantic.Field(
        default=None,
        examples=[None, "images"],
        description="For running vision language model tuning pass "
        "the column name of the image data in the dataset.",
    )

    remove_unused_columns: bool | None = pydantic.Field(
        default=True,
        examples=[True, False],
        description="Remove columns not required by the model when using an nlp.Dataset.",
    )

    dataset_kwargs_skip_prepare_dataset: bool | None = pydantic.Field(
        default=False,
        examples=[True, False],
        description="When True, configures trl to skip preparing the dataset.",
    )

    flash_attn: bool = pydantic.Field(
        default=True,
        examples=[True, False],
        description="Use Flash attention v2 from transformers",
    )

    @pydantic.field_validator("peft_method", mode="before")
    def upgrade_peft_method(cls, value: str) -> str | None:
        if isinstance(value, str) and value == "full":
            return None
        return value

    @pydantic.field_validator("fast_moe", mode="before")
    def upgrade_fast_moe(cls, value: int | list[int] | None) -> list[int] | None:
        # VV: Currently, fast_moe has a single argument in it (ep_degree). It's easier to describe the property
        # domain of discrete values so we're going to assume that this is a single integer for now
        if isinstance(value, (int, float)):
            if int(value) != value:
                # VV: ado converts integers to floats - if the value has decimal digits raise an exception
                return value
            value = int(value)

            if value == 0:
                return None
            return [value]
        return value


# VV: Different experiments have different required properties. Here we keep a record of their names
# We'll use this information to partition EntitySpace into:
#  1. hardcoded parameters in the Experiment metadata (options a user cannot override at all)
#  2. optional parameters (options a user may override but is not required to provide)
#  3. required parameters (options a user must provide)

MINIMUM_PROPS = [
    "model_name",
    "model_max_length",
    "batch_size",
    "number_gpus",
    # "gpu_model",
]

DEPRECATED_MINIMUM_PROPS_SINGLE_NODE = [*MINIMUM_PROPS, "dataset_id", "torch_dtype"]

DEPRECATED_MINIMUM_PROPS_MULTI_NODE = [
    *DEPRECATED_MINIMUM_PROPS_SINGLE_NODE,
    "number_nodes",
]


class EntitySpace(SFTTrainerCLIArgs):
    """This contains Entity properties which however do not map to a CLI argument"""

    # VV: If you're updating these, then make sure you also update domain_for_constitutive_property()
    # the code uses `examples` to populate the categorical values of the constitutive property's domain

    # VV: no examples as the number of gpus depends on whether you're using FSDP, DDP, or DP
    number_gpus: int = pydantic.Field(
        examples=[0, 1, 2, 4, 8],
        description="The total number of GPUs to use",
        exclude=True,
    )

    gpu_model: str | None = pydantic.Field(
        default=None,
        examples=[
            None,
            "NVIDIA-A100-SXM4-80GB",
            "NVIDIA-A100-80GB-PCIe",
            "Tesla-T4",
            "L40S",
            "Tesla-V100-PCIE-16GB",
            "NVIDIA-H100-PCIe",
            "NVIDIA-H100-80GB-HBM3",
        ],
        description="The GPU model to use",
        exclude=True,
    )

    distributed_backend: typing.Literal["DDP", "FSDP"] | None = pydantic.Field(
        default="FSDP",
        examples=["DDP", "FSDP", "None"],
        description="Which pytorch backend to use when training with multiple GPU devices",
        exclude=True,
    )

    number_nodes: int = pydantic.Field(
        1,
        examples=[1, 2, 3, 4],
        description="If set, actuator distributes tasks on multiple nodes. "
        "Each Node will use number_gpus/number_nodes GPUs. "
        "Each Node will use 1 process for each GPU it uses",
        exclude=True,
    )

    fms_hf_tuning_version: str | None = pydantic.Field(
        # VV: In #1175 we decided that we're not going to update the default value of this experiment
        default="2.1.2",
        examples=list(FMS_HF_TUNING_VERSION),
        description="The version of fms-hf-tuning to use - controls which wrapper to use "
        "as well as python dependencies",
        exclude=True,
    )

    enable_roce: bool = pydantic.Field(
        False,
        examples=[False, True],
        description="This setting is only in effect for multi-node runs. "
        "It controls whether RDMA over Converged Ethernet (RoCE) is switched on or not",
        exclude=True,
    )

    # VV: TAG: @HF_RAM_Efficient_Training
    # VV: These are settings we need to replicate the huggingface post we store them in DistributedSettings
    fsdp_sharding_strategy: typing.Literal[
        "FULL_SHARD", "SHARD_GRAD_OP", "NO_SHARD", "HYBRID_SHARD", "HYBRID_SHARD_ZERO2"
    ] = pydantic.Field(
        default="FULL_SHARD",
        description="[1] FULL_SHARD (shards optimizer states, gradients and parameters), "
        "[2] SHARD_GRAD_OP (shards optimizer states and gradients), "
        "[3] NO_SHARD (DDP), "
        "[4] HYBRID_SHARD (shards optimizer states, gradients and parameters within each node "
        "while each node has full copy), "
        "[5] HYBRID_SHARD_ZERO2 (shards optimizer states and gradients within each node while each node has "
        "full copy). For more information, please refer the official PyTorch docs.",
        examples=[
            "FULL_SHARD",
            "SHARD_GRAD_OP",
            "NO_SHARD",
            "HYBRID_SHARD",
            "HYBRID_SHARD_ZERO2",
        ],
        exclude=True,
    )

    fsdp_state_dict_type: typing.Literal[
        "FULL_STATE_DICT", "LOCAL_STATE_DICT", "SHARDED_STATE_DICT"
    ] = pydantic.Field(
        default="FULL_STATE_DICT",
        description="[1] FULL_STATE_DICT, [2] LOCAL_STATE_DICT, [3] SHARDED_STATE_DICT",
        examples=["FULL_STATE_DICT", "LOCAL_STATE_DICT", "SHARDED_STATE_DICT"],
        exclude=True,
    )

    fsdp_use_orig_params: bool = pydantic.Field(
        default=True,
        description="If True, allows non-uniform `requires_grad` during init, which means support for "
        "interspersed frozen and trainable parameters. (useful only when `use_fsdp` flag is passed).",
        exclude=True,
        examples=[False, True],
    )

    accelerate_config_mixed_precision: typing.Literal["no", "fp16", "bf16", "fp8"] = (
        pydantic.Field(
            default="no",
            examples=["no", "fp16", "bf16", "fp8"],
            description="Whether or not to use mixed precision training. Choose from 'no', 'fp16', 'bf16' or 'fp8'. "
            "'fp8' requires the installation of transformers-engine.",
            exclude=True,
        )
    )

    accelerate_config_fsdp_transformer_layer_cls_to_wrap: str | None = pydantic.Field(
        default=None,
        examples=[
            None,
            "GraniteDecoderLayer",
            "LlamaDecoderLayer",
            "MistralDecoderLayer",
            "GPTJBlock",
            "T5Block",
        ],
        description=(
            "List of transformer layer class names (case-sensitive) to wrap, e.g, BertLayer, "
            "GraniteDecoderLayer, GPTJBlock, T5Block ... (useful only when using FSDP)"
        ),
        exclude=True,
    )

    @pydantic.field_validator("dataset_id")
    def val_dataset_id(cls, value: str) -> str:
        if value not in DatasetMap:
            raise ValueError("Unknown dataset")

        return value

    @classmethod
    def domain_for_constitutive_property(
        cls,
        property: str,
        field_info: pydantic.fields.FieldInfo,
    ) -> "PropertyDomain | None":
        # VV: number_gpus need to be divisible by fast_moe so we can assume that they have the same range
        # also, both fast_moe and number_gpus can be set to 0
        gpus_range = [0, 32 + 1]

        discrete_properties_with_range = {
            "number_gpus": gpus_range,
            "batch_size": [1, 4096 + 1],
            "model_max_length": [1, (2 << 16) + 1],
            "gradient_accumulation_steps": [1, 32 + 1],
            "number_nodes": [1, 8 + 1],
            "r": [1, 32 + 1],
            "lora_alpha": [1, 32 + 1],
            "fast_moe": gpus_range,
        }

        if property in discrete_properties_with_range:
            return PropertyDomain(
                domainRange=discrete_properties_with_range[property],
                interval=1,
                variableType=VariableTypeEnum.DISCRETE_VARIABLE_TYPE,
            )

        values = field_info.examples

        if not values:
            raise ValueError(
                f"EntitySpace.{property} has no examples in the BaseModel - fix it"
            )

        if isinstance(values[0], bool):
            variable_type = VariableTypeEnum.BINARY_VARIABLE_TYPE
        elif isinstance(values[0], (int, float)):
            variable_type = VariableTypeEnum.DISCRETE_VARIABLE_TYPE
        else:
            variable_type = VariableTypeEnum.CATEGORICAL_VARIABLE_TYPE

        return PropertyDomain(
            values=values,
            variableType=variable_type,
        )

    @pydantic.field_validator("peft_method", mode="before")
    def upgrade_peft_method(cls, value: str | None) -> str | None:
        if isinstance(value, str) and value == "full":
            return None

        return value

    @pydantic.field_validator(
        "distributed_backend",
        mode="before",
    )
    def set_none_string_to_none_literal(cls, value: str | None) -> str | None:
        if isinstance(value, str) and value == "None":
            return None

        return value

    @classmethod
    def orch_experiment_required_properties(
        cls,
    ) -> "list[ConstitutiveProperty]":

        return [
            ConstitutiveProperty(
                identifier=identifier,
                metadata={"description": field_info.description},
                propertyDomain=cls.domain_for_constitutive_property(
                    property=identifier, field_info=field_info
                ),
            )
            for identifier, field_info in cls.model_fields.items()
        ]

    def validate_and_update(
        self, exp_params: "ExperimentParameters", logger: logging.Logger
    ):
        """Method updates both @self and @exp_params to make the 2 consistent with each other, also validates them

        Args:
            exp_params:
                The experiment-specific parameters. May get updated in place
            logger:
                A logger object

        Raises:
            InvalidEntityError:
                When the experiment parameters or entity space contain invalid information
        """

        # VV: Here we convert the "effective" batch size to the per_device_train_batch_size.
        # Unfortunately, some values of batch_size and number_gpus are describing inconsistent runs, for example
        # batch_size = 1 and number_gpus = 2 would mean that 2 devices are expected to process 1 batch_size.
        # We could either artificially increase the batch_size to 2, reduce the number_gpus or just report
        # that this entity is not valid. Here, we opt for the latter.
        if self.batch_size % max(1, self.number_gpus) != 0:
            # VV: This marks the run as `success` - there will be just 1 measured property: `is_valid = 0`
            raise InvalidEntityError(
                f"batch_size is {self.batch_size} but number_gpus is {self.number_gpus}"
            )

        if exp_params.multi_node is not None and (
            self.number_nodes > 1 and not exp_params.multi_node
        ):
            raise InvalidEntityError(
                f"number_nodes is {self.number_nodes} but experiment is single node"
            )

        if self.distributed_backend is not None and self.number_gpus < 2:
            logger.info(
                f"Requested an {self.distributed_backend} experiment using {self.number_gpus} "
                f"gpus will set distributed_backend to None"
            )
            self.distributed_backend = None

        if self.number_nodes == 1 and exp_params.multi_node:
            logger.info(
                "Requested a multi_node experiment but number_nodes == 1, "
                "will run the equivalent single node experiment instead"
            )
            exp_params.multi_node = False

        if self.number_nodes > 1:
            exp_params.multi_node = True

        if max(self.number_gpus, 1) % self.number_nodes != 0:
            raise InvalidEntityError(
                f"number_gpus is {self.number_gpus} but number_nodes is {self.number_nodes}"
            )

        if self.fast_moe:
            ep_degrees = self.fast_moe[0]

            if self.number_gpus % ep_degrees != 0:
                raise InvalidEntityError(
                    f"number_gpus is {self.number_gpus} but fast_moe[0] (i.e. ep_degrees) is {ep_degrees}"
                )

            import ado_actuators.sfttrainer.wrapper_fms_hf_tuning.tuning_versions as tv

            if (
                not self.fms_hf_tuning_version
                or tv.semver_cmp(tv.semver_parse(self.fms_hf_tuning_version), (2, 4, 0))
                == -1
            ):
                raise InvalidEntityError(
                    "fast_moe is supported for fms-hf-tuning version v2.4.0 and onwards but "
                    f"fms_hf_tuning_version is {self.fms_hf_tuning_version}"
                )


# VV: Get the defaults from the definition of the EntitySpace
DEFAULT_PARAMS = {
    k: v.default
    for k, v in EntitySpace.model_fields.items()
    if not isinstance(v.default, pydantic_core._pydantic_core.PydanticUndefinedType)
}


class ExperimentParameters(pydantic.BaseModel):
    # VV: Fields which do not map to a CLI arg of tuning.sft_trainer should come with exclude=True

    model_config = pydantic.ConfigDict(
        extra="forbid", protected_namespaces=(), use_enum_values=True
    )

    multi_node: bool | None = pydantic.Field(
        None,
        description="Set for experiments which can only do one of multi-node or single-gpu measurements",
        exclude=True,
    )
    # VV: For example 4.0
    num_train_epochs: float = 1.0
    max_steps: int = -1

    gradient_checkpointing: bool

    # VV: Describes the goals of this experiment (e.g. measure performance metrics, get stability statistics, etc)
    purpose: ExperimentPurpose = pydantic.Field(
        ExperimentPurpose.Performance,
        description="What this experiment is aiming to measure",
        exclude=True,
    )

    # VV: Different experiments expect different weight formats
    weights_format: WeightsFormat = pydantic.Field(
        description="The kind of weights this experiment can load",
        exclude=True,
    )

    def args_for_entity_space(
        self, entity_space: EntitySpace, model_map: dict[WeightsFormat, str]
    ) -> dict[str, typing.Any]:

        args = self.model_dump()
        try:
            args["model_name"] = model_map[self.weights_format]
        except KeyError:
            raise NotImplementedError(
                f"Experiment expects weights_format {self.weights_format} but "
                f"the available weight formats for the model are {list(model_map)}"
            )
        return args


class PromptTuningExperimentParameters(ExperimentParameters):
    peft_method: typing.Literal["pt"] = pydantic.Field(
        description="This is a prompt-tuning experiment"
    )


class FullFinetuningExperimentsParameters(ExperimentParameters):
    peft_method: None = pydantic.Field(
        description="This is a full fine-tuning experiment"
    )

    @pydantic.field_validator("peft_method", mode="before")
    def upgrade_peft_method(cls, value: str) -> str | None:
        if isinstance(value, str) and value == "full":
            return None
        return value


class LoraExperimentParameters(ExperimentParameters):
    peft_method: typing.Literal["lora"] = pydantic.Field(
        description="This is a LORA experiment"
    )

    # r: int
    #
    # lora_alpha: int

    target_modules_map: dict[str, list[str]] = pydantic.Field(
        description="A map of model_name to list of target_modules for LORA",
        exclude=True,
    )

    def args_for_entity_space(
        self, entity_space: EntitySpace, model_map: dict[WeightsFormat, str]
    ) -> dict[str, typing.Any]:
        args = super().args_for_entity_space(
            entity_space=entity_space, model_map=model_map
        )
        try:
            args["target_modules"] = self.target_modules_map[entity_space.model_name]

        except KeyError:
            raise NotImplementedError(
                f"No target_modules mapping for model {entity_space.model_name}"
            )

        return args


class GPTQLoraExperimentParameters(LoraExperimentParameters):
    auto_gptq: str

    fp16: str

    fast_kernels: list[str]

    fused_lora: list[str]

    torch_dtype: typing.Literal["float16"]

    # VV: Anh Uong said that padding_free is not required as GPTQ-LoRA was tested before support for
    # padding_free was available
    # padding_free: typing.List[str]

    def args_for_entity_space(
        self, entity_space: EntitySpace, model_map: dict[WeightsFormat, str]
    ) -> dict[str, typing.Any]:

        if entity_space.torch_dtype != "float16" and self.auto_gptq == "triton_v2":
            raise InvalidEntityError("triton_v2 only supports torch_dtype=float16")

        return super().args_for_entity_space(
            entity_space=entity_space, model_map=model_map
        )


def experiment_parameters_from_experiment(
    exp: "Experiment",
    entity_values: dict[str, typing.Any],
) -> ExperimentParameters:
    import json

    method = exp.metadata.get("method")

    parameters = exp.metadata.get("parameters") or {}

    try:
        model_class: type[pydantic.BaseModel] = {
            "pt": PromptTuningExperimentParameters,
            "lora": LoraExperimentParameters,
            "full": FullFinetuningExperimentsParameters,
            "full-error": FullFinetuningExperimentsParameters,
            "gptq-lora": GPTQLoraExperimentParameters,
        }[method]
    except KeyError:
        raise InternalInconsistencyError(
            f"There is no ExperimentParameters schema for experiment {exp.identifier} with method {method}"
        )

    # VV: Copy in any fields that the experiment sets and override the defaults from the experiment metadata
    # this also fills in any missing fields from the experiment metadata
    if entity_values:
        for name, value in entity_values.items():
            if name in model_class.model_fields:
                parameters[name] = value

    try:
        return model_class.model_validate(parameters)
    except Exception:
        print(
            "instantianting",
            exp.identifier,
            model_class.__name__,
            "with",
            json.dumps(parameters),
        )
        raise



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import json
import os.path

import ray
import yaml


@ray.remote
def patch_weights():
    # VV: Last time I checked, the llama 7b gptq weights on the s3 bucket are broken
    path = "/hf-models-pvc/LLaMa/models/hf/7B-gptq/quantize_config.json"
    if not os.path.isfile(path):
        with open(path, "w") as f:
            json.dump(
                {
                    "bits": 4,
                    "group_size": 128,
                    "damp_percent": 0.01,
                    "desc_act": 0,
                    "static_groups": False,
                    "sym": True,
                    "true_sequential": True,
                    "model_name_or_path": None,
                    "model_file_base_name": None,
                    "is_marlin_format": False,
                    "quant_method": "gptq",
                },
                f,
            )


@ray.remote
def download(
    uri: str, dest: str, secret_access_key: str, access_key_id: str, endpoint: str
):
    import subprocess
    import sys

    env = os.environ.copy()

    env.update(
        {
            "AWS_SECRET_ACCESS_KEY": secret_access_key,
            "AWS_ACCESS_KEY_ID": access_key_id,
        }
    )

    command = ["aws", "s3", "--endpoint", endpoint, "sync", uri, dest]

    print(command)

    proc = subprocess.Popen(command, stdout=sys.stderr, stderr=sys.stderr, env=env)
    proc.wait()

    if proc.returncode != 0:
        raise ValueError(f"Download failed with returncode {proc.returncode}")


@ray.remote
def upload(
    src: str, uri: str, secret_access_key: str, access_key_id: str, endpoint: str
):
    import subprocess
    import sys

    env = os.environ.copy()

    env.update(
        {
            "AWS_SECRET_ACCESS_KEY": secret_access_key,
            "AWS_ACCESS_KEY_ID": access_key_id,
        }
    )

    command = ["aws", "s3", "--endpoint", endpoint, "sync", src, uri]

    print(command)

    proc = subprocess.Popen(command, stdout=sys.stderr, stderr=sys.stderr, env=env)
    proc.wait()

    if proc.returncode != 0:
        raise ValueError(f"Upload failed with returncode {proc.returncode}")


def main():
    ray.init()

    warning = """The file {path} should look like this:
bucket: fmaas-integration-tests
endpoint: https://s3.us-east.cloud-object-storage.appdomain.cloud/
access_key_id: <something>
secret_access_key: <something>

The credentials are in the `SFTTrainer` vault on 1Password, if you don't have access to it
ask Vassilis Vassiliadis or Alessandro Pomponio to grant you access"""

    try:
        with open("fm-openshift_credentials.yaml") as f:
            fm_openshift_credentials = yaml.safe_load(f)
    except Exception:
        print(warning.format(path="fm-openshift_credentials.yaml"))
        raise

    try:
        with open("fmas-integration-tests-credentials.yaml") as f:
            fmas_integration_tests_credentials = yaml.safe_load(f)
    except Exception:
        print(warning.format(path="fmas-integration-tests-credentials.yaml"))
        raise

    # VV: These are the models we download from a place other than huggingface
    fm_openshift_paths = {
        "/hf-models-pvc/LLaMa/models/hf/llama2-70b/": "models/external/llama2/70B/",
        "/hf-models-pvc/LLaMa/models/hf/llama3-70b/": "models/external/llama3/70b_pre_trained/",
        "/hf-models-pvc/granite-8b-code-base/": "models/ibm-granite-hf/granite-8b-code-base/",
        "/hf-models-pvc/granite-34b-code-base/": "models/ibm-granite-hf/granite-34b-code-base/",
        "/hf-models-pvc/Mixtral-8x7B-Instruct-v0.1/": "models/external/mistralai/Mixtral-8x7B-Instruct-v0.1/",
    }

    # VV: These are all the GPTQ weights
    # VV: format is {"destination in pvc": "source in bucket"}
    fmas_integration_tests_paths = {
        "/hf-models-pvc/granite-34b-gptq/": "models/granite-34b-gptq/",
        "/hf-models-pvc/LLaMa/models/hf/llama3-70b-gptq/": "models/llama3-70b-gptq/",
        "/hf-models-pvc/mixtral_8x7b_instruct_v0.1_gptq": "models/mixtral_8x7b_instruct_v0.1_gptq/",
        "/hf-models-pvc/mistral-7B-v0.3-gptq": "models/mistral-7B-v0.3-gptq/",
        "/hf-models-pvc/granite-20b-code-base-v2/step_280000_ckpt-gptq/": "models/granite-20b-gptq/",
        "/hf-models-pvc/LLaMa/models/hf/7B-gptq/": "models/llama-2-7b-gptq/",
        "/hf-models-pvc/granite-8b-code-instruct-gptq/": "models/granite-8b-code-instruct-gptq/",
        "/hf-models-pvc/LLaMa/models/hf/llama3.1-405b-gptq": "models/llama3-405b-gptq/",
        "/hf-models-pvc/granite-7b-base-gtpq/": "models/granite_7b_base_gptq/",
    }

    for creds, models in [
        (fmas_integration_tests_credentials, fmas_integration_tests_paths),
        (fm_openshift_credentials, fm_openshift_paths),
    ]:
        for dest, src in models.items():
            print(f"Download {dest} from {src}")
            if not dest.endswith("/"):
                dest += "/"
            if not src.endswith("/"):
                src += "/"

            ray.get(
                download.remote(
                    uri=os.path.join("s3://", creds["bucket"], src),
                    dest=dest,
                    access_key_id=creds["access_key_id"],
                    secret_access_key=creds["secret_access_key"],
                    endpoint=creds["endpoint"],
                )
            )

            if dest == "/hf-models-pvc/LLaMa/models/hf/7B-gptq/":
                ray.get(patch_weights.remote())


if __name__ == "__main__":
    main()



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

from ado_actuators.sfttrainer.ray_env.utils import (
    apply_exclude_package_rules,
    packages_requiring_nvidia_development_binaries,
)


def test_exclude_packages():
    packages = [
        "yarl==1.20.0",
        "bitsandbytes==0.43.3",
        "causal-conv1d==1.5.0.post8",
        "fms-hf-tuning @ file:///wheel.whl",
        "nvidia-cublas-cu12==12.1.3.1",
        "flash_attn==2.7.4.post1",
    ]
    exclude_packages = [
        *packages_requiring_nvidia_development_binaries(),
        "bitsandbytes",
        "fms-hf-tuning",
    ]
    filtered, dropped = apply_exclude_package_rules(exclude_packages, packages)

    assert filtered == ["yarl==1.20.0"]
    assert dropped == [
        "bitsandbytes==0.43.3",
        "causal-conv1d==1.5.0.post8",
        "fms-hf-tuning @ file:///wheel.whl",
        "nvidia-cublas-cu12==12.1.3.1",
        "flash_attn==2.7.4.post1",
    ]



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

"""
# Generates a synthetic vision dataset from a seed file

## Documentation

The Seed file should be a human-readable text file. This script will parse the file and use it to build N entries of
a dataset. N is the product of num_max_batch_size * num_max_gpus * num_gradient_accumulation_steps

The motivation behind the formula of N is that the dataset should be large enough to support 1 full epoch for the
maximum considered: batch size, number of gpus, and number of gradient accumulation steps. If this property of the
dataset is violated, then the metrics that SFTTrainer from transformers produces can be inaccurate.

Each dataset entry is crafted such that:

1. it has about 50 tokens as its input (*)
2. it has about 18200 tokens (by default) as its output (*)
3. both input and output tokens consist of streams of consecutive words found in the seed file (**)
4. each sample contains a 384x384 image with a couple of geometric shapes in it
5. each sample consists of a conversation between a user and an assistant. The user asks a question (input tokens)
   and provides an image. The assistant provides an answer (output tokens)
6. the images are embedded in the dataset as streams of bytes

## Caveats

- (*) The script assumes that each token is roughly 0.75 "words" (space separated text). The script keeps the count of
    the words it uses for each entry in the dataset and uses 0.75 words per token rule to estimate the number of
    tokens in that entry. This estimation is just a "good enough" guess. For example, different models may use
    different tokenizers and as a result they may produce a different number of tokens.
- (**) The script starts using the "words" from the beginning of the file and may wrap around the file if necessary.
    It keeps track of where it stopped so that different entries of the dataset do not all start from the exact same
    part of the seed file.

## Implications of the above

1. The synthetic dataset is dependent on the seed file you use. By default, the script uses itself as its seed file.
   My understanding is that 2 different seed files will produce "computationally equivalent" synthetic datasets.
   However, we have not formally verified this assumption.
2. When you load the synthetic dataset with your tokenizer, you will observe that the tokens for each entry in the
   dataset are actually different from what you requested. This is normal behaviour. We recommend asking for more
   tokens than you think you will need and controlling the actual tokens that your model trains on by setting the
   parameter --max_seq_length of fms-hf-tuning.
"""

import io
import logging
import os.path
import pathlib
import sys

import pandas.io.parquet
import typer
from PIL import Image, ImageDraw

app = typer.Typer(rich_markup_mode="markdown")


def generate_image(width: int = 384, height: int = 384) -> bytes:
    image = Image.new("RGB", (width, height), "white")
    draw = ImageDraw.Draw(image)

    x = width // 2
    y = height // 2
    radius = 200

    # VV: Draw 2 circles in the image
    draw.ellipse(
        (
            x - radius,
            y - radius,
            x + radius,
            y + radius,
        ),
        fill="red",
    )

    draw.ellipse(
        (
            x - radius * 0.5,
            y - radius * 0.5,
            x + radius * 0.5,
            y + radius * 0.5,
        ),
        fill="blue",
    )

    buffer = io.BytesIO()
    image.save(buffer, format="JPEG")
    return buffer.getvalue()


def generate_sequence(words: list[str], start: int, length: int) -> tuple[str, int]:
    ret = []
    idx = 0
    while idx < length:
        batch = min(length - idx, len(words) - start)

        ret.extend(words[start : start + batch])
        idx += batch

        start = (start + batch) % len(words)

    return " ".join(ret), start


def compute_population_size(
    num_max_batch_size=128, num_max_gpus=8, num_gradient_accumulation_steps=4
) -> int:
    return num_max_batch_size * num_max_gpus * num_gradient_accumulation_steps


def generate_dataset(
    seed_file: pathlib.Path,
    image: bytes,
    tokens_input: int = 10,
    tokens_output: int = 20000,
    words_per_token: float = 0.75,
    population: int = 4096,
) -> pandas.DataFrame:
    with open(seed_file, encoding="utf-8") as f:
        words = [x for x in f.read().split() if len(x) > 0]

    prompts = []
    dataset = {
        "images": [{"bytes": image, "path": None}] * population,
        "output": prompts,
    }
    start = 0

    words_prompt = int(tokens_input * words_per_token + 0.5)
    words_response = int(tokens_output * words_per_token + 0.5)

    for idx in range(population):
        text_input, start = generate_sequence(words, start, words_prompt)
        text_output, start = generate_sequence(words, start, words_response)

        entry = [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": text_input, "index": None},
                    {"type": "image", "index": idx},
                ],
            },
            {
                "role": "assistant",
                "content": [{"type": "text", "text": text_output, "index": None}],
            },
        ]

        prompts.append(entry)

    return pandas.DataFrame(dataset)


@app.command(
    no_args_is_help=True,
    help="Generates a synthetic dataset which should have at least @tokens_input+@tokens_output tokens in "
    "each entry. You can use the same dataset in experiment runs with different batch sizes, number of gpus, "
    "gradient accumulation steps. You just need to generate one dataset that contains enough entries "
    "for a full epoch of your largest experiment.",
    epilog=f"Examples\n\n"
    f"- python {sys.argv[0]} -o dataset.jsonl\n\n"
    f"- python {sys.argv[0]} -i path/to/some/file -o dataset.jsonl",
)
def main(
    output: pathlib.Path = typer.Option(
        ...,
        "--output",
        "-o",
        help="Where to store the dataset file.",
    ),
    input: pathlib.Path = typer.Option(
        pathlib.Path(__file__),
        "--input",
        "-i",
        exists=True,
        file_okay=True,
        dir_okay=False,
        help="Input file to use as a source of sentences. For example you can use this python file, "
        "or some kind of publicly available document such as the Apache 2.0 License file. "
        "Smaller files should also work just as well.",
    ),
    tokens_input: int = typer.Option(
        50,
        help="How many tokens to have before the Response delimiter",
    ),
    tokens_output: int = typer.Option(
        18200,
        help="How many tokens to have after the Response delimiter",
    ),
    words_per_token: float = typer.Option(
        0.75,
        help="How many tokens you expect to have per token - this is just an estimation",
    ),
    num_max_gpus: int = typer.Option(
        8, help="Maximum number of gpus in your experiment campaign"
    ),
    num_max_batch_size: int = typer.Option(
        128, help="Maximum batch size in your experiment campaign"
    ),
    num_gradient_accumulation_steps: int = typer.Option(
        4,
        help="Maximum num_gradient_accumulation_steps you will investigate with your experiment campaign",
    ),
    image_width: int = typer.Option(384, help="The image width in pixels"),
    image_height: int = typer.Option(384, help="The image height in pixels"),
    log_level: int = typer.Option(20, "--log-level", "-l", help="Log level"),
):
    logging.basicConfig(
        level=log_level,
        format="%(levelname)-9s %(name)-30s: %(funcName)-20s %(asctime)-15s: %(message)s",
    )
    population = compute_population_size(
        num_max_gpus=num_max_gpus,
        num_max_batch_size=num_max_batch_size,
        num_gradient_accumulation_steps=num_gradient_accumulation_steps,
    )
    logging.info(f"Generating the dataset with {population} entries")
    ds = generate_dataset(
        input,
        image=generate_image(image_width, image_height),
        tokens_input=tokens_input,
        tokens_output=tokens_output,
        words_per_token=words_per_token,
        population=population,
    )

    logging.info(f"Saving file under {output}")
    if not os.path.isdir(output.parent.as_posix()):
        os.makedirs(output.parent.as_posix(), exist_ok=True)
    pandas.io.parquet.to_parquet(ds, output)
    logging.info("Done")


if __name__ == "__main__":
    app()



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import json
import logging
import os

import ado_actuators.sfttrainer.wrapper_fms_hf_tuning.finetune as finetune
import ray

logging.basicConfig(level=10)

"""Computes statistics about the number of tokens per entry in the dataset.
It also populates the num_tokens_cache_dir with cached files containing the tokens per entry of a dataset
based on the tokenizer of a model

example output:

{
    "news-tokens-16384plus-entries-4096": {
        "llama3.1-8b": {
            "min": 17238,
            "max": 31959,
            "avg": 22518.065185546875
        },
        "llama3.1-70b": {
            "min": 17238,
            "max": 31959,
            "avg": 22518.065185546875
        },
        "llama3.1-405b": {
            "min": 17238,
            "max": 31959,
            "avg": 22518.065185546875
        }
    },
    "news-tokens-128kplus-entries-4096": {
        "llama3.1-8b": {
            "min": 154084,
            "max": 154434,
            "avg": 154241.513671875
        },
        "llama3.1-70b": {
            "min": 154084,
            "max": 154434,
            "avg": 154241.513671875
        },
        "llama3.1-405b": {
            "min": 154084,
            "max": 154434,
            "avg": 154241.513671875
        }
    }
}
"""


@ray.remote(
    runtime_env={
        "env_vars": {
            "LOG_LEVEL": "debug",
            "LOGLEVEL": "debug",
            "HF_HOME": "/hf-models-pvc/huggingface_home",
        },
    },
)
def tokenize_text(
    path_model: str,
    path_data: str,
    model_id: str | None,
    num_tokens_cache_dir: str | None,
):
    num_tokens = finetune.get_tokens_per_sample_in_dataset(
        path_model=path_model,
        path_data=path_data,
        model_id=model_id,
        num_tokens_cache_dir=num_tokens_cache_dir,
        dataset_text_field="output",
    )

    sum_tokens = sum(num_tokens)

    ret = {
        "min": min(num_tokens),
        "max": max(num_tokens),
        "avg": sum_tokens / len(num_tokens),
        "ds_file": os.path.splitext(os.path.basename(path_data))[0],
        "model_id": model_id,
    }

    logger = logging.getLogger("sft_trainer")
    logger.info(json.dumps(ret))

    return ret


def main():
    ray.init()
    root_data = os.environ.get("DATA_PATH", "/data/fms-hf-tuning/artificial-dataset/")

    # {dataset_id: {model_id: sum_tokens}}
    dataset_sizes = {}

    dataset_files = [
        # "news-chars-512-entries-4096.jsonl",
        # "news-chars-1024-entries-4096.jsonl",
        # "news-chars-2048-entries-4096.jsonl",
        "news-tokens-16384plus-entries-4096.jsonl",
    ]

    large_dataset_files = {
        "news-tokens-16384plus-entries-4096.jsonl",
        "news-tokens-128kplus-entries-4096.jsonl",
    }

    small_models = {
        "llama-7b": "/hf-models-pvc/LLaMa/models/hf/7B",
        "granite-13b-v2": "/hf-models-pvc/granite-13b-base-v2/step_300000_ckpt",
        "llama-13b": "/hf-models-pvc/LLaMa/models/hf/13B",
        "granite-20b-v2": "/hf-models-pvc/granite-20b-code-base-v2/step_280000_ckpt/",
        "granite-7b-base": "ibm-granite/granite-7b-base",
        "granite-8b-japanese": "/hf-models-pvc/granite-8b-japanese-base-v1-llama/",
        "granite-8b-code-base": "/hf-models-pvc/granite-8b-code-base/",
        "granite-34b-code-base": "/hf-models-pvc/granite-34b-code-base/",
        "mistral-7b-v0.1": "/hf-models-pvc/mistralai-mistral-7b-v0.1",
        "llama3-8b": "/hf-models-pvc/LLaMa/models/hf/llama3-8b",
        "llama3-70b": "/hf-models-pvc/LLaMa/models/hf/llama3-70b/",
        "mixtral-8x7b-instruct-v0.1": "/hf-models-pvc/Mixtral-8x7B-Instruct-v0.1/",
        "llama2-70b": "/hf-models-pvc/LLaMa/models/hf/llama2-70b/",
    }

    large_models = {
        "llama3.1-8b": "/hf-models-pvc/LLaMa/models/hf/llama3.1-8b",
        "llama3.1-70b": "/hf-models-pvc/LLaMa/models/hf/llama3.1-70b",
        "llama3.1-405b": "/hf-models-pvc/LLaMa/models/hf/llama3.1-405b",
    }

    cache_dir = os.path.join(root_data, "cache")

    os.makedirs(cache_dir, exist_ok=True)

    tasks = []

    for dataset_collection, models in [
        (dataset_files, small_models),
        (dataset_sizes, large_models),
        (large_dataset_files, large_models),
    ]:
        for ds_file in dataset_collection:
            path_data = os.path.join(root_data, ds_file)

            for model_id, path_model in models.items():
                tasks.append(
                    tokenize_text.remote(
                        path_model=path_model,
                        path_data=path_data,
                        model_id=model_id,
                        num_tokens_cache_dir=cache_dir,
                    )
                )

    results = ray.get(tasks)

    for ret in results:
        if ret["ds_file"] not in dataset_sizes:
            dataset_sizes[ret["ds_file"]] = {}

        dataset_sizes[ret["ds_file"]][ret["model_id"]] = {
            k: v for k, v in ret.items() if k not in ("model_id", "ds_file")
        }

    print(json.dumps(dataset_sizes, indent=2))


if __name__ == "__main__":
    main()



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import copy
import typing

import orchestrator.schema.property
from orchestrator.modules.actuators.catalog import ExperimentCatalog

from . import common

if typing.TYPE_CHECKING:
    import orchestrator.modules.actuators.catalog


def add_full_experiments(
    catalog: "orchestrator.modules.actuators.catalog.ExperimentCatalog",
):
    method = "full"
    version = "1.0.0"
    exp_name = f"finetune_{method}_benchmark"

    description = (
        "Measures the performance of full-finetuning a model for a given "
        "(GPU model, number GPUS, batch_size, model_max_length, number nodes) combination."
    )

    hardcoded_parameters: dict[str, typing.Any] = {
        "peft_method": method,
        "weights_format": common.WeightsFormat.Vanilla,
        "purpose": common.ExperimentPurpose.Performance,
    }

    # VV: Here configure any propertyDomains which differ from the default ones in EntitySpace
    override_propertydomains = {}

    default_params = copy.deepcopy(common.DEFAULT_PARAMS)
    param_experiment = common.generate_parameterisable_finetune_experiment(
        hardcoded_parameters=hardcoded_parameters,
        default_params=default_params,
        override_propertydomains=override_propertydomains,
        version=version,
        method=method,
        description=description,
        exp_identifier=f"{exp_name}-v{version}",
        exp_name=exp_name,
        actuator_identifier=common.ACTUATOR_IDENTIFIER,
        fms_hf_tuning_versions=[".".join([str(d) for d in v]) for v in common.semvers],
        required_property_names=common.MINIMUM_PROPS,
    )

    catalog.addExperiment(param_experiment)


def add_full_stability_experiments(catalog: "ExperimentCatalog"):

    method = "full"
    version = "1.0.0"
    purpose = common.ExperimentPurpose.Stability
    exp_name = f"finetune_{method}_{purpose.lower()}"

    description = (
        "Performs 5 full finetune runs of 5 steps each on a model "
        "and reports the fraction of those that resulted in GPU OOM, Other error, "
        "or No Error for a given (GPU model, number GPUS, batch_size, model_max_length) combination."
    )

    hardcoded_parameters: dict[str, typing.Any] = {
        "max_steps": 5,
        "weights_format": common.WeightsFormat.Vanilla,
        "peft_method": None,
        "purpose": purpose,
    }

    # VV: Here configure any propertyDomains which differ from the default ones in EntitySpace
    override_propertydomains = {}

    default_params = {
        x: v
        for x, v in copy.deepcopy(common.DEFAULT_PARAMS).items()
        if x not in hardcoded_parameters
    }
    param_experiment = common.generate_parameterisable_finetune_experiment(
        hardcoded_parameters=hardcoded_parameters,
        default_params=default_params,
        override_propertydomains=override_propertydomains,
        version=version,
        method=method,
        description=description,
        exp_identifier=f"{exp_name}-v{version}",
        exp_name=exp_name,
        actuator_identifier=common.ACTUATOR_IDENTIFIER,
        fms_hf_tuning_versions=[".".join([str(d) for d in v]) for v in common.semvers],
        required_property_names=common.MINIMUM_PROPS,
        properties=[
            "f_gpu_oom",
            "f_other_error",
            "f_no_error",
            "is_valid",
        ],
    )

    catalog.addExperiment(param_experiment)


def add_experiments(catalog: "ExperimentCatalog"):
    add_full_experiments(catalog)
    add_full_stability_experiments(catalog)



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import copy
import typing

from orchestrator.modules.actuators.catalog import ExperimentCatalog

from . import common

default_target_modules = {
    "smollm2-135m": ["q_proj", "v_proj"],
    "granite-3b-code-base-128k": ["q_proj", "v_proj"],
    "granite-7b-base": ["q_proj", "v_proj"],
    "granite-8b-code-base-128k": ["q_proj", "v_proj"],
    "granite-8b-code-base": ["q_proj", "v_proj"],
    "granite-8b-japanese": ["q_proj", "v_proj"],
    "granite-13b-v2": ["c_attn", "c_proj"],
    "granite-20b-v2": ["c_attn", "c_proj"],
    "granite-34b-code-base": ["c_attn", "c_proj"],
    "llama-7b": ["q_proj", "k_proj"],
    "llama-13b": ["q_proj", "k_proj"],
    "llama2-70b": ["q_proj", "v_proj"],
    "llama3-8b": ["q_proj", "k_proj"],
    "llama3-70b": ["q_proj", "v_proj"],
    "llama3.1-8b": ["q_proj", "v_proj"],
    "llama3.2-1b": ["q_proj", "v_proj"],
    "llama3.2-3b": ["q_proj", "v_proj"],
    "llama3.1-70b": ["q_proj", "v_proj"],
    "llama3.1-405b": ["q_proj", "v_proj"],
    "allam-1-13b": ["q_proj", "v_proj"],
    "hf-tiny-model-private/tiny-random-BloomForCausalLM": [
        "dense_h_to_4h",
        "dense_4h_to_4h",
    ],
    "mistral-7b-v0.1": ["q_proj", "v_proj"],
    "mistral-123b-v2": ["q_proj", "v_proj"],
    "mixtral-8x7b-instruct-v0.1": ["q_proj", "v_proj"],
    "granite-3-8b": ["q_proj", "v_proj"],
    "granite-3.1-2b": ["q_proj", "v_proj"],
    "granite-3.1-8b-instruct": ["q_proj", "v_proj"],
    "granite-4.0-micro": ["q_proj", "v_proj"],
    "granite-4.0-h-1b": ["q_proj", "v_proj"],
    "granite-4.0-350m": ["q_proj", "v_proj"],
    "granite-4.0-h-small": ["q_proj", "v_proj"],
    "granite-4.0-h-micro": ["q_proj", "v_proj"],
    "granite-4.0-h-tiny": ["q_proj", "v_proj"],
    "granite-3.0-1b-a400m-base": ["q_proj", "v_proj"],
    "granite-3.1-3b-a800m-instruct": ["q_proj", "v_proj"],
    "granite-vision-3.2-2b": ["q_proj", "v_proj"],
    "llava-v1.6-mistral-7b": ["q_proj", "v_proj"],
}


def add_experiments(catalog: "ExperimentCatalog"):

    method = "lora"
    version = "1.0.0"
    exp_name = f"finetune_{method}_benchmark"

    description = (
        "Measures the performance of LORA tuning a model for a given "
        "(GPU model, number GPUS, batch_size, model_max_length, number nodes) combination."
    )

    hardcoded_parameters: dict[str, typing.Any] = {
        "peft_method": method,
        "weights_format": common.WeightsFormat.Vanilla,
        "target_modules_map": default_target_modules,
        "purpose": common.ExperimentPurpose.Performance,
    }

    # VV: an EntityBase can represent many different kind of measurements.
    # This experiment is ONLY for LORA measurements
    override_propertydomains = {}

    default_params = copy.deepcopy(common.DEFAULT_PARAMS)
    default_params.update({"r": 4, "lora_alpha": 16})
    param_experiment = common.generate_parameterisable_finetune_experiment(
        hardcoded_parameters=hardcoded_parameters,
        default_params=default_params,
        override_propertydomains=override_propertydomains,
        version=version,
        method=method,
        description=description,
        exp_identifier=f"{exp_name}-v{version}",
        exp_name=exp_name,
        actuator_identifier=common.ACTUATOR_IDENTIFIER,
        fms_hf_tuning_versions=[".".join([str(d) for d in v]) for v in common.semvers],
        required_property_names=common.MINIMUM_PROPS,
    )

    catalog.addExperiment(param_experiment)



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

"""Uses the artificial dataset containing input and output strings consisting of 512 characters to produce
a dataset whose input string is a dataset of 512 characters and output contains (512+2) * 9 + 512 characters.
Then it converts this dataset to the alpaca format.

File structure
- this_file.py
- common_en_news_combined_1024.jsonl
"""

# Standard
import os

# Third Party
import datasets


def format_alpaca_fn(example):
    ex_input = example["input"]
    ex_output = ". ".join([example["output"] for _ in range(100)])

    return {"output": f"### Input:\n{ex_input}\n\n### Response: {ex_output}"}


num_max_batch_size = 128
num_max_gpus = 8
num_gradient_accumulation = 4

enhanced_size = num_max_batch_size * num_max_gpus * num_gradient_accumulation


path_dir = os.path.abspath(os.path.dirname(__file__))
path_input = os.path.join(path_dir, "common_en_news_combined_1024.jsonl")

path_output = os.path.join(
    path_dir, f"news-tokens-16384plus-entries-{enhanced_size}.jsonl"
)

ds: datasets.dataset_dict.DatasetDict = datasets.load_dataset(
    "json", data_files=path_input
)

alpaca_ds: datasets.arrow_dataset.Dataset = ds["train"].map(
    format_alpaca_fn, remove_columns=["input"]
)

x = alpaca_ds.to_list()

replicated = [x[i % len(x)] for i in range(enhanced_size)]

alpaca_ds = datasets.arrow_dataset.Dataset.from_list(replicated)

alpaca_ds.to_json(path_output)



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

"""Preprocesses some of the artificial news dataset

File structure
- this_file.py
- common_en_news_combined_512.jsonl
- common_en_news_combined_1024.jsonl
- common_en_news_combined_512.jsonl
"""

# Standard
import os

# Third Party
import datasets


def format_alpaca_fn(example):
    prompt = (
        "Write a response that appropriately completes the remainder of the following input text.\n\n"
        "### Input:\n{input}\n\n### Response: {output}"
    )
    output = prompt.format_map(example)
    return {"output": output}


num_max_batch_size = 128
num_max_gpus = 8
num_gradient_accumulation = 4

enhanced_size = num_max_batch_size * num_max_gpus * num_gradient_accumulation

for num in [512, 1024, 2048]:
    path_dir = os.path.abspath(os.path.dirname(__file__))
    path_input = os.path.join(path_dir, f"common_en_news_combined_{num}.jsonl")

    path_output = os.path.join(
        path_dir, f"news-chars-{num}-entries-{enhanced_size}.jsonl"
    )

    ds: datasets.dataset_dict.DatasetDict = datasets.load_dataset(
        "json", data_files=path_input
    )

    alpaca_ds: datasets.arrow_dataset.Dataset = ds["train"].map(
        format_alpaca_fn, remove_columns=["input"]
    )

    x = alpaca_ds.to_list()

    replicated = [x[i % len(x)] for i in range(enhanced_size)]

    alpaca_ds = datasets.arrow_dataset.Dataset.from_list(replicated)

    alpaca_ds.to_json(path_output)



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import dataclasses
import datetime
import logging
import os
import typing
from threading import Event, Thread

import aim.ext.resource.stat
import psutil

if typing.TYPE_CHECKING:
    from tuning.sft_trainer import EnhancedTrainOutput


def was_gpu_in_use(
    gpu_cuda_id: int, cuda_visible_devices: list[int], num_gpus: int
) -> bool:
    # VV: We used that GPU if cuda_visible_devices is empty BUT we did ask
    # FSDP to use at least 1 GPU. OR the gpu cuda id is in the cuda_visible_devices array (i.e. AIM/pytorch knew
    # exactly which of the GPUs available on the system it was meant to use)
    return (not cuda_visible_devices and num_gpus > 0) or (
        gpu_cuda_id in cuda_visible_devices
    )


@dataclasses.dataclass
class AggregatedValues:
    avg: float | None = dataclasses.field(default=-1.0)
    min: float | None = dataclasses.field(default=-1.0)
    max: float | None = dataclasses.field(default=-1.0)


def aggregate_values(
    values: list[float] | dict[str, float | None],
) -> float | AggregatedValues:
    if isinstance(values, list):
        len_values = 0
        _sum = 0
        avg = None
        _min = None
        _max = None

        for x in values:
            if x is None:
                continue
            len_values += 1
            _sum += x

            if _min is None or _min > x:
                _min = x
            if _max is None or _max < x:
                _max = x

        if len_values > 0:
            avg = _sum / len_values

        values = {
            "avg": avg,
            "max": _max,
            "min": _min,
        }

    return AggregatedValues(**values)


@dataclasses.dataclass
class GPUMetrics:
    gpu: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues, metadata={"help": "GPU utilization percent"}
    )
    gpu_memory_percent: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues,
        metadata={"help": "GPU memory utilization percent"},
    )
    gpu_temp: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues, metadata={"help": "GPU temperature"}
    )
    gpu_power_watts: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues,
        metadata={"help": "GPU power consumption in Watts."},
    )
    gpu_power_percent: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues,
        metadata={"help": "GPU power consumption (percentage)"},
    )

    @classmethod
    def from_aim_info_dict(cls, info: dict[str, typing.Any]) -> "GPUMetrics":
        return GPUMetrics(
            gpu=aggregate_values(info["__system__gpu"]["values"]),
            gpu_memory_percent=aggregate_values(
                info["__system__gpu_memory_percent"]["values"]
            ),
            gpu_temp=aggregate_values(info["__system__gpu_temp"]["values"]),
            gpu_power_watts=aggregate_values(
                info["__system__gpu_power_watts"]["values"]
            ),
            gpu_power_percent=aggregate_values(
                info["__system__gpu_power_percent"]["values"]
            ),
        )


@dataclasses.dataclass
class SystemMetrics:
    cpu: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues,
        metadata={"help": "CPU utilization percent (can be over 1.0)"},
    )
    memory_percent: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues, metadata={"help": "Whole system memory usage"}
    )
    p_memory_percent: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues,
        metadata={"help": "Memory occupied by process"},
    )
    disk_percent: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues, metadata={"help": "Disk usage"}
    )

    @classmethod
    def from_aim_info_dict(cls, info: dict[str, typing.Any]) -> "SystemMetrics":
        return SystemMetrics(
            cpu=aggregate_values(info["__system__cpu"]["values"]),
            memory_percent=aggregate_values(info["__system__memory_percent"]["values"]),
            p_memory_percent=aggregate_values(
                info["__system__p_memory_percent"]["values"]
            ),
            disk_percent=aggregate_values(info["__system__disk_percent"]["values"]),
        )


@dataclasses.dataclass
class TrainingMetrics:
    train_runtime: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues
    )
    train_samples_per_second: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues
    )
    train_steps_per_second: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues
    )
    train_tokens_per_second: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues
    )
    dataset_tokens: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues
    )

    @classmethod
    def from_aim_info_list(
        cls, training_metrics: dict[str, dict[str, typing.Any]]
    ) -> "TrainingMetrics":
        if "runtime" not in training_metrics:
            return TrainingMetrics()

        return TrainingMetrics(
            train_runtime=aggregate_values(training_metrics["runtime"]["values"]),
            train_samples_per_second=aggregate_values(
                training_metrics["samples_per_second"]["values"]
            ),
            train_steps_per_second=aggregate_values(
                training_metrics["steps_per_second"]["values"]
            ),
            train_tokens_per_second=aggregate_values(
                training_metrics["tokens_per_second"]["values"]
            ),
            dataset_tokens=aggregate_values(
                training_metrics.get("dataset_tokens", {}).get("values", [])
            ),
        )


@dataclasses.dataclass
class ModelMetrics:
    model_load_time: AggregatedValues = dataclasses.field(
        default_factory=AggregatedValues
    )


@dataclasses.dataclass
class Metrics:
    gpus: list[GPUMetrics]
    system: SystemMetrics
    training: TrainingMetrics
    model: ModelMetrics
    training_steps: int = dataclasses.field(
        metadata={"help": "The number of training steps excluding the warmup steps"},
    )
    aim_run_hash: str | None = None
    train_time_start: datetime.datetime | None = None
    train_time_stop: datetime.datetime | None = None
    hostname_gpus: dict[str, list[int]] = dataclasses.field(
        default_factory=dict,
        metadata={
            "help": "Key=value pairs where the key is a hostname and the value an "
            "array of GPU indices that the host used"
        },
    )
    warmup_steps: int | None = None
    warmup_seconds: float | None = None

    def to_scalar_observations(
        self,
        distributed_backend: typing.Literal["FSDP", "DDP"] | None,
        world_size: int,
    ) -> dict[str, float]:
        scalar_observations = {}
        import dataclasses

        def aggregate_metrics(
            vals: list[float],
        ) -> tuple[float, float, float]:
            """Utility method that returns min, avg, max of a list of floats
            If the array is empty then the method assumes that it's equal to [0.0]
            """
            vals = vals or [0.0]
            return min(vals), sum(vals) / len(vals), max(vals)

        scalar_observations.update(
            {
                "cpu_compute_utilization": self.system.cpu.avg,
                "cpu_memory_utilization": self.system.memory_percent.avg,
            }
        )

        for label, values in (
            ("gpu_compute_utilization", [gpu.gpu.avg for gpu in self.gpus]),
            (
                "gpu_memory_utilization",
                [gpu.gpu_memory_percent.avg for gpu in self.gpus],
            ),
            ("gpu_power_watts", [gpu.gpu_power_watts.avg for gpu in self.gpus]),
            ("gpu_power_percent", [gpu.gpu_power_percent.avg for gpu in self.gpus]),
        ):
            v_min, v_avg, v_max = aggregate_metrics(values)
            scalar_observations[f"{label}_min"] = v_min
            scalar_observations[f"{label}_avg"] = v_avg
            scalar_observations[f"{label}_max"] = v_max

        if self.gpus:
            scalar_observations["gpu_memory_utilization_peak"] = max(
                [gpu.gpu_memory_percent.max for gpu in self.gpus],
            )

        scalar_observations.update(
            {
                k: v["avg"]
                for k, v in dataclasses.asdict(self.training).items()
                if k != "dataset_tokens"
            }
        )

        if world_size > 1 and distributed_backend in ["FSDP", "DDP"]:
            # VV: FSDP and DDP report the train_tokens_per_second per device
            scalar_observations["train_tokens_per_second"] *= world_size

        if self.training.dataset_tokens.avg is not None:
            scalar_observations["dataset_tokens_per_second"] = (
                self.training.dataset_tokens.avg / self.training.train_runtime.avg
            )
            scalar_observations["dataset_tokens_per_second_per_gpu"] = (
                scalar_observations["dataset_tokens_per_second"] / world_size
            )
        else:
            scalar_observations["dataset_tokens_per_second"] = -1.0
            scalar_observations["dataset_tokens_per_second_per_gpu"] = -1.0

        scalar_observations["train_tokens_per_gpu_per_second"] = (
            scalar_observations["train_tokens_per_second"] / world_size
        )
        scalar_observations["model_load_time"] = self.model.model_load_time.avg

        if self.aim_run_hash:
            scalar_observations["aim_run_hash"] = self.aim_run_hash

        if self.hostname_gpus:
            scalar_observations["hostname_gpus"] = self.hostname_gpus

        if self.train_time_stop is not None:
            scalar_observations["train_time_stop"] = self.train_time_stop

        if self.train_time_start is not None:
            # VV: This includes the warmup phase
            scalar_observations["train_time_start"] = self.train_time_start

        scalar_observations["training_steps"] = self.training_steps

        return scalar_observations

    @classmethod
    def from_aim_info_dict(cls, aim_info: dict[str, typing.Any], num_gpus: int):
        json_metrics = aim_info["metrics"]

        training_metrics = {
            x["name"]: x
            for x in json_metrics
            if x["name"]
            in (
                "samples_per_second",
                "steps_per_second",
                "tokens_per_second",
                "runtime",
                "dataset_tokens",
            )
        }

        training_metrics = TrainingMetrics.from_aim_info_list(training_metrics)

        gpus = {}

        for m in json_metrics:
            if m["name"].startswith("__system__gpu"):
                gpu_cuda_id = m["context"]["gpu"]

                if gpu_cuda_id not in gpus:
                    gpus[gpu_cuda_id] = {
                        # VV: init all metrics just in case aim couldn't extract any of them
                        x: {"values": [], "context": {"gpu": gpu_cuda_id}, "name": x}
                        for x in [
                            "__system__gpu",
                            "__system__gpu_temp",
                            "__system__gpu_power_watts",
                            "__system__gpu_memory_percent",
                            "__system__gpu_power_percent",
                        ]
                    }

                gpus[gpu_cuda_id][m["name"]] = m

        gpu_metrics = []
        for gpu_cuda_id, gpu_info in gpus.items():
            if was_gpu_in_use(gpu_cuda_id, aim_info["cuda_visible_devices"], num_gpus):
                gpu_metrics.append(GPUMetrics.from_aim_info_dict(gpu_info))

        system_metrics = {
            # VV: init all metrics just in case aim couldn't extract any of them
            x: {"values": [], "context": {}, "name": x}
            for x in [
                "__system__cpu",
                "__system__disk_percent",
                "__system__memory_percent",
                "__system__p_memory_percent",
            ]
        }

        for m in json_metrics:
            if m["name"] in system_metrics:
                system_metrics[m["name"]] = m

        system_metrics = SystemMetrics.from_aim_info_dict(system_metrics)

        model_load_time = [-1.0]

        for m in json_metrics:
            if m["name"] == "model_load_time":
                model_load_time = m["values"]

        train_time_start = aim_info.get("train_time_start")
        train_time_stop = aim_info.get("train_time_stop")
        format_time = "%d%m%y-%H%M%S"

        if train_time_start:
            train_time_start = datetime.datetime.strptime(train_time_start, format_time)
        if train_time_stop:
            train_time_stop = datetime.datetime.strptime(train_time_stop, format_time)

        # VV: training_steps does not include the warmup_steps
        training_steps = aim_info["training_steps"] - (
            aim_info.get("warmup_steps") or 0
        )

        warmup_seconds = aim_info.get("warmup_seconds") or 0

        if warmup_seconds > 0:
            # VV: Just like training_steps, train_runtime does not include the warmup_seconds
            training_metrics.train_runtime.avg -= warmup_seconds
            training_metrics.train_runtime.min -= warmup_seconds
            training_metrics.train_runtime.max -= warmup_seconds

        return Metrics(
            gpus=gpu_metrics,
            system=system_metrics,
            training=training_metrics,
            model=ModelMetrics(model_load_time=aggregate_values(model_load_time)),
            aim_run_hash=aim_info.get("run_hash"),
            hostname_gpus={aim_info["hostname"]: aim_info["cuda_visible_devices"]},
            train_time_start=train_time_start,
            train_time_stop=train_time_stop,
            training_steps=training_steps,
            warmup_steps=aim_info.get("warmup_steps"),
            warmup_seconds=aim_info.get("warmup_seconds"),
        )

    def filter_unused_gpus(self):
        """Uses CUDA_VISIBLE_DEVICES to filter out metrics of GPUs that the task did not use

        Updates the metrics in place
        """
        # VV: CUDA_VISIBLE_DEVICES can be a comma separated list of GPU indices that this process
        # has access to. If this is the case, we can use this information to filter the field `gpus` of metrics
        try:
            cuda_devices = os.environ["CUDA_VISIBLE_DEVICES"]
            cuda_devices = [int(x) for x in cuda_devices.split(",")]
            total_gpus = len(self.gpus)

            if len(cuda_devices) and all(x < total_gpus for x in cuda_devices):
                self.gpus = [self.gpus[idx] for idx in sorted(cuda_devices)]
        except (KeyError, ValueError):
            pass


def round10e5(val):
    return round(val * 10e5) / 10e5


class ResourceTracker:
    AGG_MODE_AVG = "average"
    AGG_MODE_MIN = "min"
    AGG_MODE_MAX = "max"
    AGG_MODE_DIFF = "diff"
    AGG_NOTHING = "nothing"
    AGG_DEFAULT = AGG_NOTHING

    @classmethod
    def aggregate(cls, items: list, mode: str):
        """
        Aggregates array of numbers by a given 'mode'
        """
        if mode == cls.AGG_MODE_MAX:
            return max(items)
        if mode == cls.AGG_MODE_MIN:
            return min(items)
        if mode == cls.AGG_MODE_AVG:
            return round10e5(sum(items) / len(items))
        if mode == cls.AGG_MODE_DIFF:
            return round10e5(max(items) - min(items))
        if mode == cls.AGG_NOTHING:
            return items
        raise ValueError(f"unknown aggregation mode: '{mode}'")

    def __init__(self, period: float = 30.0):
        """Takes a snapshot of system metrics every @period seconds

        Args:
            period:
                Seconds between successive snapshots
        """
        self._log = logging.getLogger("ResourceTracker")
        self._process = psutil.Process()
        self._stat = aim.ext.resource.stat.Stat(self._process)
        self._fix_stat(self._stat.stat_item)

        self.metrics: list[aim.ext.resource.stat.StatDict] = [self._stat.stat_item]

        # VV: end_track() puts a message here asking the thread to stop
        self._end_requested = Event()
        self._stopped = Event()

        self._thread: Thread | None = None
        self._stopped.set()
        self._period = period

    def track_stats(self):
        system, gpus = self._stat.get_stats()
        stat = aim.ext.resource.stat.StatDict(system, gpus)

        try:
            self._fix_stat(stat)
        finally:
            self._log.debug(f"Resources Snapshot {stat.to_dict()}")

        self.metrics.append(stat)

    def _kernel(self: "ResourceTracker"):
        """Utility method which takes a metric snapshot every period seconds and can be cancelled via end_track()"""

        while self._end_requested.wait(timeout=self._period) is False:
            self.track_stats()

        # VV: ensure at least 1 measurement but instead of taking it at the start of the measurements
        # track the metrics right after the method was asked to shutdown. This way we don't
        # capture the state of machine "before" the training started
        self.track_stats()

        self._stopped.set()

    def begin_track(self):
        self._stopped.clear()

        self._thread = Thread(target=self._kernel)
        self._thread.start()

    def end_track(self):
        if self._thread is not None:
            self._log.debug("Stopping stat tracker thread")
            self._end_requested.set()
            self._thread.join()
            # VV: Just some sanity checking
            assert self._stopped.is_set()
            self._thread = None

    @classmethod
    def _fix_stat(cls, stat: aim.ext.resource.stat.StatDict):
        # VV: There's a bug in AIM which causes some GPU fields to be a tuple of 1 item instead of a float
        for g in stat.gpus:
            for name in g:
                if isinstance(g[name], tuple):
                    if len(g[name]) != 1:
                        raise ValueError(f"Unexpected GPU stat {name} in GPU {g}")
                    g[name] = g[name][0]

    @classmethod
    def aggregate_items(
        cls,
        items: list[aim.ext.resource.stat.StatDict],
        agg_mode: str = AGG_NOTHING,
    ):
        """
        Aggregates array of `StatDict` items by a given `mode`
        """
        aggregated_stat = aim.ext.resource.stat.StatDict()

        # Return empty item if items array is empty
        if not items or len(items) == 0:
            return aggregated_stat

        gpu_stats = []
        for s in items:
            # Collect system stats
            for k in s.system:
                aggregated_stat.system.setdefault(k, [])
                aggregated_stat.system[k].append(s.system[k])

            # Collect GPU device stats
            for stat_item_gpu_idx in range(len(s.gpus)):
                stat_item_gpu_stat = s.gpus[stat_item_gpu_idx]
                if len(gpu_stats) == stat_item_gpu_idx:
                    gpu_stats.append({})
                for gpu_stat_key in stat_item_gpu_stat:
                    gpu_stat = stat_item_gpu_stat[gpu_stat_key]
                    gpu_stats[stat_item_gpu_idx].setdefault(gpu_stat_key, [])
                    gpu_stats[stat_item_gpu_idx][gpu_stat_key].append(gpu_stat)

        # Aggregate system stats
        for k in aggregated_stat.system:
            aggregated_stat.system[k] = cls.aggregate(
                aggregated_stat.system[k], agg_mode
            )

        # Aggregate GPU device stats
        for g in range(len(gpu_stats)):
            for k in gpu_stats[g]:
                gpu_stats[g][k] = cls.aggregate(gpu_stats[g][k], agg_mode)

        aggregated_stat.gpus = gpu_stats

        return aggregated_stat

    def to_metrics(self, train_output: "EnhancedTrainOutput | None") -> Metrics:
        self._log.debug("Waiting for stat tracker thread to finish")
        self._stopped.wait()

        aggregate = self.aggregate_items(self.metrics, agg_mode=self.AGG_NOTHING)
        gpus = [
            GPUMetrics(**{k: aggregate_values(v) for k, v in g.items()})
            for g in aggregate.gpus
        ]

        system = SystemMetrics(
            **{k: aggregate_values(v) for k, v in aggregate.system.items()}
        )
        train_metrics = {}
        model_load_time = -1

        if train_output is not None:
            model_load_time = train_output.model_load_time
            train_metrics.update(
                {
                    k.name: train_output.train_output.metrics[k.name]
                    for k in dataclasses.fields(TrainingMetrics)
                    if k.name in train_output.train_output.metrics
                }
            )
        training = TrainingMetrics(**train_metrics)

        model = ModelMetrics(model_load_time=aggregate_values([model_load_time]))
        return Metrics(gpus=gpus, system=system, training=training, model=model)

    @classmethod
    def filter_unused_gpus(cls, metrics: Metrics):
        """Uses CUDA_VISIBLE_DEVICES to filter out metrics of GPUs that the task did not use

        Args:
            metrics:
                The metrics to update in place
        """
        metrics.filter_unused_gpus()



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import contextlib
import functools
import os
import typing


@functools.cache
def is_using_arm_cpu() -> bool:
    """Returns True when device is using an ARM cpu"""
    import platform

    platform_machine = platform.machine().lower()

    return platform_machine == "aarch64" or "arm" in platform_machine


@functools.cache
def is_nvcc_available() -> bool:
    """Returns True when nvcc is in $PATH"""
    import shutil

    return shutil.which("nvcc") is not None


@functools.cache
def is_uv_available() -> bool:
    """Returns True when uv is in $PATH"""
    import shutil

    return shutil.which("uv") is not None


@functools.cache
def is_nvidia_smi_available() -> bool:
    """Returns True when nvidia-smi is in $PATH"""
    import shutil

    return shutil.which("nvidia-smi") is not None


@functools.cache
def is_pip_available() -> bool:
    """Returns True when pip is import-able"""
    import importlib.util

    spec = importlib.util.find_spec("pip")
    return spec is not None and spec.loader is not None


# VV: Need to include the ray_runtime_env_plugins_string parameter here for caching to work
@functools.cache
def _check_if_ray_will_load_ordered_pip_plugin(
    ray_runtime_env_plugins: str | None,
) -> bool:
    import json

    from orchestrator.utilities.ray_env.ordered_pip import OrderedPipPlugin

    if not is_pip_available():
        return False

    with contextlib.suppress(Exception):
        decoded = json.loads(ray_runtime_env_plugins)

        for entry in decoded:
            if (
                isinstance(entry, dict)
                and "class" in entry
                and entry["class"] == OrderedPipPlugin.ClassPath
            ):
                return True

    return False


def is_ordered_pip_available() -> bool:
    """Returns True when ray is configured to load the ordered_pip RuntimeEnvPlugin and pip is importable"""
    return _check_if_ray_will_load_ordered_pip_plugin(
        os.environ.get("RAY_RUNTIME_ENV_PLUGINS")
    )


def packages_requiring_nvidia_development_binaries():
    return [
        "fms-acceleration-foak",
        "fms-acceleration-moe",
        "triton",
        "flash_attn",
        "mamba-ssm",
        "causal-conv1d",
        # VV: mamba_ssm and causal_conv1d changed their package names
        "mamba_ssm",
        "causal_conv1d",
        "nvidia-cublas-cu12",
        "nvidia-cuda-cupti-cu12",
        "nvidia-cuda-nvrtc-cu12",
        "nvidia-cuda-runtime-cu12",
        "nvidia-cudnn-cu12",
        "nvidia-cufft-cu12",
        "nvidia-cufile-cu12",
        "nvidia-curand-cu12",
        "nvidia-cusolver-cu12",
        "nvidia-cusparse-cu12",
        "nvidia-cusparselt-cu12",
        "nvidia-nccl-cu12",
        "nvidia-nvjitlink-cu12",
        "nvidia-nvshmem-cu12",
        "nvidia-nvtx-cu12",
    ]


def apply_exclude_package_rules(
    exclude_packages: list[str], packages: list[str]
) -> tuple[list[str], list[str]]:
    """Filters out packages based on a list of exclusion rules.

    Args:
        exclude_packages:
            A list of rules for excluding package names
        packages:
            A list of packages to filter

    Returns:
        A tuple with 2 items.
            The first item is the list containing only the packages that do not match any of the exclusion rules
            The second item is the list of packages that were removed
    """
    if not exclude_packages:
        return packages, []

    ret = []
    removed = []

    for candidate_package in packages:
        # VV: Some packages look like this: "name @ file://...."
        trimmed = candidate_package.replace(" ", "")
        for unwanted_package in exclude_packages:
            if (
                trimmed.startswith((f"{unwanted_package}=", f"{unwanted_package}@"))
                or trimmed == unwanted_package
            ):
                removed.append(candidate_package)
                break
        else:
            # VV: Keep the original package
            ret.append(candidate_package)

    return ret, removed


def get_pinned_packages(
    path_requirements: str,
    override_fms_hf_tuning: str | None = None,
    ensure_aim: bool = True,
    exclude_packages: list[str] | None = None,
) -> list[str]:
    """Extracts the pinned packages from a path_requirements file

    Args:
        path_requirements:
            Path to the requirements.txt file containing the pinned packages (one package per line, a-la pip)
        override_fms_hf_tuning:
            If set, overrides the `fms-hf-tuning` pinned package from the contents of the requirements.txt file
        ensure_aim:
            If set, ensures that the dependencies include the aim python package.
        exclude_packages:
            Packages to exclude from installation
    Returns:
        An array consisting of pinned packages a-la pip
    """

    with open(path_requirements, encoding="utf-8") as f:
        packages = [x.strip() for x in f if x.strip() and not x.startswith("#")]

    def find_matching_packages(package_name: str, packages: list[str]) -> list[str]:
        return [
            x
            for x in packages
            if x.startswith((f"{package_name}=", f"{package_name} "))
            or x == package_name
        ]

    if override_fms_hf_tuning:
        exclude_packages = [*list(exclude_packages or []), "fms-hf-tuning"]

    packages, _dropped = apply_exclude_package_rules(exclude_packages, packages)

    if override_fms_hf_tuning:
        packages.append(override_fms_hf_tuning)

    if ensure_aim and len(find_matching_packages("aim", packages)) == 0:
        packages.append("aim")

    return packages


@functools.cache
def ray_version_supports_pip_install_options() -> bool | None:
    import ray

    try:
        # VV: Ray added support for pip_install_options in 2.50.0
        version = [int(x) for x in ray.__version__.split(".")]
        return version[0] > 2 or (version[0] >= 2 and version[1] >= 50)
    except Exception as e:
        print(
            f"Unable to tell whether pip_install_options is available for Ray Runtime environments due to {e!s} - "
            f"will assume that it is unavailable"
        )
        return None


def packages_depending_on_torch() -> list[str]:
    # VV: mamba_ssm and causal_conv1d changed their package names
    return ["flash_attn", "mamba-ssm", "causal-conv1d", "mamba_ssm", "causal_conv1d"]


def get_ray_environment(
    packages: list[str],
    packages_requiring_extra_phase: list[list[str]],
    env_vars: dict[str, str],
) -> dict[str, typing.Any]:
    """Builds a ray-environment using a Ray RuntimeEnvPlugin.

    The function picks the RuntimeEnvPlugin by inspecting the host machine and virtual environment.
    It selects the plugins using the following priority:

    1. ordered_pip
    2. pip
    3. uv

    Args:
        packages:
            The list of python packages to install
        packages_requiring_extra_phase:
            A list of lists of packages. The list with index i expects that the packages in the list with index i-1
            have already been installed in the virtual environment that ray will be building.
            This is only used when the ordered_pip RuntimeEnvPlugin is available. Otherwise, it is ignored.
        env_vars:
            Environment variables to inject into the RuntimeContext
    Returns:
        A dictionary representing a RuntimeContext for Ray jobs
    """
    if is_ordered_pip_available():
        env_plugin_name = "ordered_pip"
    elif is_pip_available():
        env_plugin_name = "pip"
    elif is_uv_available():
        env_plugin_name = "uv"
    else:
        raise NotImplementedError(
            "No uv binary in $PATH, pip cannot be imported, and ordered_pip is not configured. "
            "Ensure your virtual environment is valid."
        )

    # VV: Do not switch on pip_check.
    plugin = {}
    env = {"AIM_UI_TELEMETRY_ENABLED": "0"}
    env.update(env_vars)

    ray_environment = {
        "env_vars": env,
        env_plugin_name: plugin,
    }

    pip_install_options = []

    if ray_version_supports_pip_install_options() or env_plugin_name == "uv":
        # VV: Ray added support for pip_install_options in 2.50.0
        pip_install_options = ["--no-build-isolation"]

        if env.get("PIP_FIND_LINKS"):
            # VV: I find that exporting PIP_FIND_LINKS does not behave the same way as using --find-links
            pip_install_options.extend(("--find-links", env["PIP_FIND_LINKS"]))

    if env_plugin_name == "pip":
        phase = {"packages": packages}

        if pip_install_options:
            phase["pip_install_options"] = pip_install_options
        else:
            ray_environment["env_vars"]["PIP_NO_BUILD_ISOLATION"] = "0"

        plugin.update(phase)
    elif env_plugin_name == "uv":
        pip_install_options.insert(0, "--no-build-isolation")
        plugin.update(
            {"uv_pip_install_options": pip_install_options, "packages": packages}
        )
    elif env_plugin_name == "ordered_pip":
        # VV: Keeps the linter happy
        base_packages = []
        phases = [{"packages": base_packages}]
        plugin["phases"] = phases

        for p in packages_requiring_extra_phase or []:
            packages, this_phase = apply_exclude_package_rules(
                exclude_packages=p, packages=packages
            )
            if this_phase:
                phase = {"packages": this_phase}
                if ray_version_supports_pip_install_options():
                    phase["pip_install_options"] = list(pip_install_options)
                phases.append(phase)

        # VV: At this point the packages var contains all the packages that must go into the very first phase
        base_packages.extend(packages)
    else:
        raise NotImplementedError("Unknown ray environment env plugin", env_plugin_name)

    return ray_environment



# Copyright (c) IBM Corporation
# SPDX-License-Identifier: MIT

import json
import os
import sys
import traceback
import typing

import ray
from ado_actuators.sfttrainer.experiments.common import (
    ModelMap,
    WeightsFormat,
)

"""Checks whether models are load-able
"""


@ray.remote(
    # resources={"Tesla-V100-PCIE-16GB": 1},
    runtime_env={
        "pip": ["accelerate", "transformers>=4.40.0"],
        "env_vars": {
            "LOG_LEVEL": "debug",
            "LOGLEVEL": "debug",
            "HF_HOME": "/hf-models-pvc/huggingface_home",
        },
    },
)
def get_model_hash(path_model: str):
    from accelerate import init_empty_weights
    from transformers import AutoConfig, AutoModelForCausalLM, AutoTokenizer

    try:
        if os.path.exists(path_model) is False:
            import huggingface_hub

            print(
                "MODEL MAY NOT EXIST WILL TRY TO DOWNLOAD IT FROM HUGGINGFACE",
                file=sys.stderr,
            )
            huggingface_hub.snapshot_download(path_model)

        with init_empty_weights():
            model_config = AutoConfig.from_pretrained(
                path_model, local_files_only=True, low_cpu_mem_usage=True
            )
            _model = AutoModelForCausalLM.from_config(model_config)
            _tokenizer = AutoTokenizer.from_pretrained(path_model)
    except Exception as e:
        print("Unable to load weights, due to", e)
        print(traceback.format_exc())
        return False

    return True


def main():
    ray.init()

    model_information: dict[str, dict[str, typing.Any]] = {}

    # VV: if you just added a model to `ModelMap` just copy/paste that entire dictionary here
    # ModelMap = { ..... }
    # (the version of the fms-hf-tuning Orchestrator Plugin on the ray cluster won't have the changes you just made)

    all_models = {
        k: v[WeightsFormat.Vanilla]
        for k, v in ModelMap.items()
        if WeightsFormat.Vanilla in v
    }

    for model_name, path_model in all_models.items():
        print(model_name, path_model)
        ret = ray.get(get_model_hash.remote(path_model))

        model_information[model_name] = ret
        print(model_name, "\n  ", json.dumps(ret))

        print("All so far")
        print(json.dumps(model_information, indent=2))

    missing = {k: False for k, v in model_information.items() if v is False}
    existing = {k: True for k, v in model_information.items() if v is True}

    print("Missing")
    print(json.dumps(missing, indent=2))

    print("Existing")
    print(json.dumps(existing, indent=2))


if __name__ == "__main__":
    main()



# Copyright The IBM Tuning Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import datetime
import os
import sys
import typing
from typing import Any

import ado_actuators.sfttrainer.wrapper_fms_hf_tuning.constants as constants
import ado_actuators.sfttrainer.wrapper_fms_hf_tuning.tuning_versions as tuning_versions
import aim
import torch
import torch.distributed
from aim.hugging_face import AimCallback
from transformers import TrainerControl, TrainerState, TrainingArguments


def has_gathered_enough_samples(
    duration_of_optimization_steps: list[float],
    warmup_seconds: float,
    meaningful_samples_seconds: float,
    meaningful_samples_amount: int,
) -> tuple[bool, int]:
    """Determines if the current process should stop based on accumulated steps and time.

    Args:
        duration_of_optimization_steps:
            A list of durations of each optimization step.
        warmup_seconds:
            The duration of the warmup period.
        meaningful_samples_seconds:
            The minimum duration required for meaningful samples.
        meaningful_samples_amount:
            The minimum number of meaningful samples required.

    Returns:
        A tuple containing a boolean indicating if the process should stop and the index of the first step after the warmup period.
        - The boolean value is True if the process should stop, False otherwise.
        - The index indicates the position of the first step after the warmup period in the original 'steps' list. -1 if the stopping criteria are not met.
    """
    if not duration_of_optimization_steps:
        return False, -1

    import numpy as np

    cs = np.cumsum(duration_of_optimization_steps)
    running_for = cs[-1]

    if running_for <= warmup_seconds + meaningful_samples_seconds:
        return False, -1

    first_after_warmup = int(np.searchsorted(cs, warmup_seconds, side="right")) + 1

    if first_after_warmup >= len(duration_of_optimization_steps):
        # VV: we need at least 1 more step
        return False, -1

    steps_after_warmup = duration_of_optimization_steps[first_after_warmup:]
    duration_of_meaningful_samples = cs[-1] - cs[first_after_warmup - 1]

    # Stop when there are at least @meaningful_samples_amount samples post AND
    # we've spent at least @meaningful_samples_seconds seconds drawing these meaningful samples
    return (
        (
            (len(steps_after_warmup) >= meaningful_samples_amount)
            and duration_of_meaningful_samples >= meaningful_samples_seconds
        ),
        first_after_warmup,
    )


def get_cuda_uuid_to_index() -> dict[str, int]:
    """Returns a dictionary mapping GPU device UUIDs to their index numbers"""
    try:
        import aim.ext.pynvml as nvml

        nvml.nvmlInit()
    except Exception as e:
        print(
            f"Unable to initialize nvml when mapping cuda uuid to AIM gpu indices due to {e} - "
            f"will skip mapping the uuids"
        )
        return {}

    gpu_device_count = nvml.nvmlDeviceGetCount()

    ret = {
        str(nvml.nvmlDeviceGetUUID(nvml.nvmlDeviceGetHandleByIndex(i))): i
        for i in range(gpu_device_count)
    }

    nvml.nvmlShutdown()

    return ret


def get_cuda_device_indices(cuda_visible_devices: str) -> list[int]:
    """Returns the indices of cuda devices

    Args:
        cuda_visible_devices: The value of the CUDA_VISIBLE_DEVICES environment
        variable. It represents the devices that should be made visible to the
        current process.

    Returns:
        a list of integers representing the device indices that should be made visible to
        the current process.
    """
    if not cuda_visible_devices:
        return []

    try:
        return [int(x) for x in cuda_visible_devices.split(",") if len(x) > 0]
    except ValueError:
        # VV: these are cuda device UIDs, need to decode them
        pass

    cuda_mapping = get_cuda_uuid_to_index()
    return [cuda_mapping.get(uuid, uuid) for uuid in cuda_visible_devices.split(",")]


def calculate_gpu_power_percent(
    run_metrics: list[tuple[str, dict[str, int], list[float]]],
):
    """Calculates __system__gpu_power_percent using __system__gpu_power_watts and inserts it into existing run metrics

    Args:
        run_metrics:
            The run metrics collected from AIM. The method updates this array in memory

    Returns:
        Nothing
    """
    from aim.ext.resource.utils import round10e5

    try:
        import aim.ext.pynvml as nvml

        nvml.nvmlInit()
    except Exception as e:
        print(
            f"Unable to instantiate nvml due to {e} - will not record power measurements",
            file=sys.stderr,
        )
        return []

    for name, context, values in run_metrics:
        # VV: aim reports: gpu_info['gpu_power_watts'] = round10e5(nvml.nvmlDeviceGetPowerUsage(handle) / 1000)
        if name == "__system__gpu_power_watts" and "gpu" in context:
            handle = nvml.nvmlDeviceGetHandleByIndex(context["gpu"])
            # VV: nvmlDeviceGetEnforcedPowerLimit is in Milliwatts:
            # https://docs.nvidia.com/deploy/nvml-api/group__nvmlDeviceQueries.html

            power_cap = nvml.nvmlDeviceGetEnforcedPowerLimit(handle) / 1000

            # VV: The range is [0, 100]
            gpu_percent = [round10e5(v * 100.0 / power_cap) for v in values]

            run_metrics.append(
                (
                    "__system__gpu_power_percent",
                    context.copy(),
                    gpu_percent,
                )
            )

    nvml.nvmlShutdown()

    return run_metrics


class CustomAimCallback(AimCallback):
    # VV: Set this after training starts and never delete it
    the_run_hash = None
    the_experiment: "aim.Run | None" = None
    training_steps = 0

    def __init__(
        self,
        repo: str | None = None,
        experiment: str | None = None,
        system_tracking_interval: int | None = 10,
        log_system_params: bool | None = True,
        capture_terminal_logs: bool | None = True,
        additional_metrics: dict[str, Any] | None = None,
        aim_info_path: str | None = None,
        aim_info_aggregate_metrics: bool = False,
        aim_metadata: dict[str, Any] | None = None,
        stop_after_seconds: float = -1.0,
        auto_stop_method: constants.AutoStopMethod | None = None,
    ):

        self._additional_metrics = additional_metrics or {}
        self._aim_info_path = aim_info_path
        self._aim_info_aggregate_metrics = aim_info_aggregate_metrics
        self._aim_metadata = aim_metadata or {}

        self._stop_after_seconds = stop_after_seconds
        self._time_started: datetime.datetime | None = None

        self._optimization_step_started: datetime.datetime | None = None

        # VV: When auto_stop_method = WARMUP_60S_STABLE_120S_OR_10_STEPS, this ivar will (eventually) be the
        # index of the first optimization step that follows the WARMUP phase.
        self._warmup_steps: int | None = None
        self._optimization_step_durations = []
        self._auto_stop_method = auto_stop_method

        super().__init__(
            repo,
            experiment,
            system_tracking_interval,
            log_system_params,
            capture_terminal_logs,
        )

    def on_train_begin(self, args, state, control, model=None, **kwargs):
        super().on_train_begin(args, state, control, model, **kwargs)

        self._time_started = datetime.datetime.now()

        if state.is_local_process_zero:
            run: aim.Run = self.experiment
            CustomAimCallback.the_experiment = run
            CustomAimCallback.the_run_hash = run.hash

            for k, v in self._additional_metrics.items():
                run.track(v, name=k, context={"scope": "additional_metrics"})

            for k, v in self._aim_metadata.items():
                run[k] = v

    def on_step_begin(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        self._optimization_step_started = datetime.datetime.now()

    def on_step_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        super().on_step_end(args, state, control, **kwargs)

        CustomAimCallback.training_steps += 1

        now = datetime.datetime.now()
        dt = (now - self._optimization_step_started).total_seconds()
        running_for = (now - self._time_started).total_seconds()

        if state.is_world_process_zero:
            self.experiment.track(value=dt, name="optimization_step_duration")

        if (
            state.is_local_process_zero
            and 0.0 <= self._stop_after_seconds <= running_for
        ):
            print(
                "Triggering experiment to stop after running for",
                running_for,
                f"seconds due to stop_after_seconds={self._stop_after_seconds}",
            )
            control.should_training_stop = True

        if state.is_local_process_zero:
            if (
                self._auto_stop_method
                == constants.AutoStopMethod.WARMUP_60S_STABLE_120S_OR_10_STEPS
            ):
                # VV: In multi-node runs, each local-rank 0 process dumps a JSON file with metrics it collects.
                # The auto_stop_method ignores system metrics gathered during the warmup phase. To achieve this,
                # each local-rank 0 process must know the warmup duration. Instead of synchronizing across workers,
                # we just have all local-rank 0 processes compute it independently.
                # To this end, local-rank 0 processes track the duration of optimization steps.
                self._optimization_step_durations.append(dt)

                # VV: all processes with local rank 0 need to compute their respective warmup phase in order to
                # be able to discard the system metrics that they collect during warmup
                gathered_enough, post_warmup_step_idx = has_gathered_enough_samples(
                    duration_of_optimization_steps=self._optimization_step_durations,
                    warmup_seconds=60,
                    meaningful_samples_seconds=120,
                    meaningful_samples_amount=10,
                )

                if gathered_enough:
                    self._warmup_steps = post_warmup_step_idx
                    print(
                        "Triggering experiment to stop after running for",
                        running_for,
                        f"seconds due to auto_stop_method={self._auto_stop_method}.",
                        "Warmup steps were",
                        self._warmup_steps,
                    )
                    control.should_training_stop = True
            elif self._auto_stop_method is not None:
                raise NotImplementedError(
                    "Unknown auto_stop_method", self._auto_stop_method
                )

        sys.stderr.flush()
        sys.stdout.flush()

        if (
            torch.distributed.is_initialized()
            and torch.distributed.get_world_size() > 1
        ):
            # VV: Stop all workers if all local-rank 0 workers agree to stop
            should_stop = torch.tensor(
                [
                    # VV: We use MIN reduction to stop training when all local-rank 0 workers agree.
                    # Non local-rank-0 workers also participate in this reduction, but their value is
                    # irrelevant - and it is always False.
                    # Here, we set these values to True so that they don't influence the reduce operation at all.
                    (not state.is_local_process_zero)
                    or control.should_training_stop
                ],
                dtype=torch.uint8,
            ).cuda()

            torch.distributed.all_reduce(should_stop, op=torch.distributed.ReduceOp.MIN)
            control.should_training_stop = should_stop.item() == 1

            if control.should_training_stop:
                print("All local-rank 0 workers agree to stop early")

    def on_train_end(self, args, state, control, **kwargs):
        try:
            if self._aim_info_path and state.is_local_process_zero:
                format_time = "%d%m%y-%H%M%S"
                train_time_stop = datetime.datetime.now().strftime(format_time)

                run: aim.Run = self.experiment

                for k, v in self._additional_metrics.items():
                    run.track(v, name=k, context={"scope": "additional_metrics"})

                cuda_visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES", "")

                cuda_visible_devices = get_cuda_device_indices(cuda_visible_devices)

                metrics = []

                run_metrics = [
                    (m.name, m.context.to_dict(), m.values.values_list())
                    for m in run.metrics()
                ]

                run_metrics.extend(calculate_gpu_power_percent(run_metrics=run_metrics))

                skip_first_system_metrics, warmup_seconds = None, None

                if self._warmup_steps:
                    warmup_seconds = sum(
                        self._optimization_step_durations[: self._warmup_steps]
                    )

                    skip_first_system_metrics = int(
                        warmup_seconds / self._system_tracking_interval
                    )

                for name, context, values in run_metrics:
                    if self._aim_info_aggregate_metrics:
                        try:
                            values = list(values)

                            if (
                                name.startswith("__system")
                                and skip_first_system_metrics is not None
                            ):
                                values = values[skip_first_system_metrics:]

                            values = [x for x in values if x is not None]
                            len_values = len(values)

                            if len_values > 0:
                                _sum = sum(values)
                                avg = _sum / len_values
                                _max = max(values)
                                _min = min(values)
                            else:
                                avg, _max, _min = None, None, None

                            values = {
                                "avg": avg,
                                "max": _max,
                                "min": _min,
                            }
                        except ValueError:
                            # Don't aggregate properties that are weird
                            pass

                    metrics.append(
                        {
                            "name": name,
                            "values": values,
                            "context": context,
                        }
                    )

                # Standard
                import json

                if self._time_started is not None:
                    train_time_start = self._time_started.strftime(format_time)
                else:
                    train_time_start = None

                with open(self._aim_info_path, "w", encoding="utf-8") as f:
                    json.dump(
                        {
                            "run_hash": run.hash,
                            "metrics": metrics,
                            "hostname": os.environ.get("HOSTNAME"),
                            "train_time_start": train_time_start,
                            "train_time_stop": train_time_stop,
                            "cuda_visible_devices": cuda_visible_devices,
                            "world_rank": os.environ.get("RANK", "0"),
                            "world_size": os.environ.get("WORLD_SIZE", "1"),
                            "training_steps": CustomAimCallback.training_steps,
                            "warmup_steps": self._warmup_steps,
                            "warmup_seconds": warmup_seconds,
                        },
                        f,
                    )
        finally:
            super().on_train_end(args=args, state=state, control=control, **kwargs)
            if CustomAimCallback.the_experiment:
                CustomAimCallback.the_experiment = None


@dataclasses.dataclass
class CustomArgs:
    aim_metadata_path: str | None = dataclasses.field(
        default=None,
        metadata={
            "help": "Path to JSON file containing metadata that sft_trainer.py will store in AIM"
        },
    )

    aim_info_path: str | None = dataclasses.field(
        default=None,
        metadata={
            "help": "The path to a JSON file that sft_trainer.py will use to store the metrics that AIM captures. "
            "If unset, the script will not produce the file"
        },
    )

    aim_info_aggregate_metrics: bool = dataclasses.field(
        default=False,
        metadata={
            "help": "Whether to store the mean values of the metrics that AIM measures in the aim_info_path file"
        },
    )

    aim_db: str | None = dataclasses.field(
        default=None,
        metadata={"help": "The AIM endpoint"},
    )
    aim_experiment: str = dataclasses.field(
        default=None,
        metadata={"help": "The name of the AIM experiment"},
    )

    fms_hf_tuning_version: str = dataclasses.field(
        default=None,
        metadata={
            "help": "The version of fms-hf-tuning to use - controls which wrapper to use "
            "as well as python dependencies"
        },
    )

    stop_after_seconds: float = dataclasses.field(
        default=-1.0,
        metadata={
            "help": "If set, the optimizer will be asked to stop after the specified time elapses. "
            "The check is performed after the end of each training step."
        },
    )

    auto_stop_method: constants.AutoStopMethod | None = dataclasses.field(
        default=None,
        metadata={
            "help": "The default value is `None`. This parameter defines the method used to automatically "
            "stop the fine-tuning job. Supported values are `WARMUP_60S_STABLE_120S_OR_10_STEPS` and "
            "`None`. If set to `WARMUP_60S_STABLE_120S_OR_10_STEPS`, the job stops after spending at least "
            "60 seconds in the warmup phase plus the longer of 120 seconds or the duration of 10 "
            "optimization steps. This method excludes the first 60 seconds of training when calculating "
            "throughput and system metrics."
        },
    )


def main():
    """Utility method that invokes the main() method of sft_trainer.py, catches GPU OOM exceptions and logs them
    to the --aim_info_path JSON file as well as STDERR"""
    import json
    import sys

    import transformers

    parser = transformers.HfArgumentParser(dataclass_types=(CustomArgs,))

    (
        custom_args,
        remaining_args,
    ) = parser.parse_args_into_dataclasses(
        args=list(sys.argv[1:]),
        return_remaining_strings=True,
    )

    sys.argv = [sys.argv[0], *remaining_args]

    custom_args = typing.cast(CustomArgs, custom_args)

    if custom_args.fms_hf_tuning_version is None:
        raise ValueError("must set --fms_hf_tuning_version")

    if custom_args.aim_metadata_path:
        with open(custom_args.aim_metadata_path) as f:
            aim_metadata = json.load(f)
    else:
        aim_metadata = {}

    import json

    import torch.cuda
    import tuning.sft_trainer

    metadata = aim_metadata.get("metadata", {})

    try:
        measurement_id = "/".join((metadata["experiment"], metadata["entity"]))
    except KeyError as e:
        print("Could not construct measurement id due to", e, file=sys.stderr)
        measurement_id = "unknown/unknown"

    def report_error(exception: Exception, warning: str, exception_type: str):
        print(warning, file=sys.stderr)
        # Standard
        import traceback

        print(traceback.format_exc(), file=sys.stderr)
        import os

        # VV: 'accelerate' injects this env-var
        if os.environ.get("RANK", ""):
            rank = os.environ["RANK"]
            path = "_".join((custom_args.aim_info_path, rank))
        else:
            path = custom_args.aim_info_path
            rank = "?"

        report = {
            "error": exception_type,
            "exception": str(exception),
            "run_hash": CustomAimCallback.the_run_hash,
            "training_steps": CustomAimCallback.training_steps,
        }

        with open(path, "w", encoding="utf-8") as f:
            json.dump(
                report,
                f,
            )

        print(
            f"Worker {rank} of {measurement_id} dumped {report} in {path}",
            file=sys.stderr,
        )

    def report_oom(exception: Exception):
        return report_error(
            exception,
            warning="SFTTRAINER_EXCEPTION: OUT_OF_MEMORY",
            exception_type="OutOfGPUMemoryError",
        )

    def report_nccl_error(exception: Exception):
        return report_error(
            exception,
            warning="SFTTRAINER_EXCEPTION: NCCL_ERROR",
            exception_type="NCCLError",
        )

    try:
        print("Worker started", file=sys.stderr)

        if not custom_args.aim_info_path:
            raise ValueError("must set --aim_info_path")

        job_config = tuning.sft_trainer.get_json_config()

        callbacks = [
            CustomAimCallback(
                repo=custom_args.aim_db,
                experiment=custom_args.aim_experiment,
                additional_metrics={},
                aim_info_path=custom_args.aim_info_path,
                aim_info_aggregate_metrics=custom_args.aim_info_aggregate_metrics,
                aim_metadata=aim_metadata,
                stop_after_seconds=custom_args.stop_after_seconds,
                auto_stop_method=custom_args.auto_stop_method,
            )
        ]
        module = tuning_versions.import_tuning_version(
            version=custom_args.fms_hf_tuning_version
        )
        module.parse_arguments_and_execute_wrapper(
            callbacks=callbacks, job_config=job_config
        )

    except torch.cuda.OutOfMemoryError as e:
        report_oom(e)
        raise
    except RuntimeError as e:
        if (
            "CUDA error: out of memory".lower() in str(e).lower()
            or "CUDA error: an illegal memory access was encountered".lower()
            in str(e).lower()
        ):
            report_oom(e)
            # elif "NCCL Error".lower() in str(e).lower():
            #     report_nccl_error(e)
        else:
            report_error(
                e,
                warning=f"SFTTRAINER_EXCEPTION: UNHANDLED {type(e)}",
                exception_type=f"Unhandled({type(e)})",
            )
        raise
    except BaseException as e:
        report_error(
            e,
            warning=f"SFTTRAINER_EXCEPTION: UNHANDLED {type(e)}",
            exception_type=f"Unhandled({type(e)})",
        )
        raise
    finally:
        print("Worker stopped", file=sys.stderr)
        if CustomAimCallback.the_experiment:
            print(
                "AIM Run is not closed, will close it now. Measurement id:",
                measurement_id,
                "AIM run hash",
                CustomAimCallback.the_experiment.hash,
                file=sys.stderr,
            )
            CustomAimCallback.the_experiment.close()


if __name__ == "__main__":
    main()
